# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import functools
import unittest
from warnings import catch_warnings

from pyface.toolkit import toolkit_object
from pyface.undo.api import AbstractCommand
from traits.api import Any, HasTraits, Int, List, Str, Tuple
from traits.testing.api import UnittestTools

from traitsui.tests.test_editor import create_editor
from traitsui.undo import AbstractUndoItem, ListUndoItem, UndoHistory, UndoItem
from traitsui.tests._tools import BaseTestMixin

GuiTestAssistant = toolkit_object("util.gui_test_assistant:GuiTestAssistant")
no_gui_test_assistant = GuiTestAssistant.__name__ == "Unimplemented"
if no_gui_test_assistant:
    # ensure null toolkit has an inheritable GuiTestAssistant
    class GuiTestAssistant(object):
        pass


class SimpleExample(HasTraits):

    value = Int()

    str_value = Str()

    tuple_value = Tuple()

    list_value = List()

    any_value = Any()


class DummyCommand(AbstractCommand):

    def do(self):
        self.data = "do"

    def undo(self):
        self.data = "undo"

    def redo(self):
        self.data = "redo"


class LegacyUndoItem(AbstractUndoItem):

    def undo(self):
        pass

    def redo(self):
        pass


class TestAbstractUndoItem(UnittestTools, unittest.TestCase):

    def test_merge_undo_deprecated(self):
        undo_item = LegacyUndoItem()
        other_item = LegacyUndoItem()

        with catch_warnings(record=True) as w:
            result = undo_item.merge(other_item)
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[0].category, DeprecationWarning))
            self.assertIn("merge_undo", str(w[0].message))

        self.assertFalse(result)


class TestUndoItem(UnittestTools, unittest.TestCase):

    def test_undo(self):
        example = SimpleExample(value=11)

        undo_item = UndoItem(
            object=example,
            name='value',
            old_value=10,
            new_value=11,
        )

        with self.assertTraitChanges(example, 'value', count=1):
            undo_item.undo()

        self.assertEqual(example.value, 10)

    def test_redo(self):
        example = SimpleExample(value=10)

        undo_item = UndoItem(
            object=example,
            name='value',
            old_value=10,
            new_value=11,
        )

        with self.assertTraitChanges(example, 'value', count=1):
            undo_item.redo()

        self.assertEqual(example.value, 11)

    def test_merge_different_undo_item_type(self):
        example_1 = SimpleExample()
        example_2 = SimpleExample()

        undo_item = UndoItem(
            object=example_1,
            name='any_value',
        )
        next_undo_item = ListUndoItem(
            object=example_2,
            name='any_value',
        )

        result = undo_item.merge(next_undo_item)

        self.assertFalse(result)

    def test_merge_different_objects(self):
        example_1 = SimpleExample()
        example_2 = SimpleExample()

        undo_item = UndoItem(
            object=example_1,
            name='value',
            old_value=10,
            new_value=11,
        )
        next_undo_item = UndoItem(
            object=example_2,
            name='value',
            old_value=10,
            new_value=11,
        )

        result = undo_item.merge(next_undo_item)

        self.assertFalse(result)

    def test_merge_different_traits(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='value',
            old_value=10,
            new_value=11,
        )
        next_undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='foo',
            new_value='bar',
        )

        result = undo_item.merge(next_undo_item)

        self.assertFalse(result)

    def test_merge_different_value_types(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='any_value',
            old_value=10,
            new_value=11,
        )
        next_undo_item = UndoItem(
            object=example,
            name='any_value',
            old_value=11,
            new_value='foo',
        )

        result = undo_item.merge(next_undo_item)

        self.assertFalse(result)

    def test_merge_numbers(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='value',
            old_value=10,
            new_value=11,
        )
        next_undo_item = UndoItem(
            object=example,
            name='value',
            old_value=11,
            new_value=12,
        )

        result = undo_item.merge(next_undo_item)

        self.assertTrue(result)
        self.assertEqual(undo_item.old_value, 10)
        self.assertEqual(undo_item.new_value, 12)

    def test_merge_str_insert(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='foo',
            new_value='bar',
        )
        next_undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='bar',
            new_value='bear',
        )

        result = undo_item.merge(next_undo_item)

        self.assertTrue(result)
        self.assertEqual(undo_item.old_value, 'foo')
        self.assertEqual(undo_item.new_value, 'bear')

    def test_merge_str_delete(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='foo',
            new_value='bear',
        )
        next_undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='bear',
            new_value='bar',
        )

        result = undo_item.merge(next_undo_item)

        self.assertTrue(result)
        self.assertEqual(undo_item.old_value, 'foo')
        self.assertEqual(undo_item.new_value, 'bar')

    def test_merge_str_change(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='foo',
            new_value='bar',
        )
        next_undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='bar',
            new_value='baz',
        )

        result = undo_item.merge(next_undo_item)

        self.assertTrue(result)
        self.assertEqual(undo_item.old_value, 'foo')
        self.assertEqual(undo_item.new_value, 'baz')

    def test_merge_str_same(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='foo',
            new_value='bar',
        )
        next_undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='bar',
            new_value='bar',
        )

        result = undo_item.merge(next_undo_item)

        self.assertTrue(result)
        self.assertEqual(undo_item.old_value, 'foo')
        self.assertEqual(undo_item.new_value, 'bar')

    def test_merge_str_different(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='foo',
            new_value='bar',
        )
        next_undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='bar',
            new_value='wombat',
        )

        result = undo_item.merge(next_undo_item)

        self.assertFalse(result)

    def test_merge_sequence_change(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='tuple_value',
            old_value=('foo', 'bar', 'baz'),
            new_value=('foo', 'wombat', 'baz'),
        )
        next_undo_item = UndoItem(
            object=example,
            name='tuple_value',
            old_value=('foo', 'wombat', 'baz'),
            new_value=('foo', 'fizz', 'baz'),
        )

        result = undo_item.merge(next_undo_item)

        self.assertTrue(result)
        self.assertEqual(undo_item.old_value, ('foo', 'bar', 'baz'))
        self.assertEqual(undo_item.new_value, ('foo', 'fizz', 'baz'))

    def test_merge_sequence_change_different_types(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='tuple_value',
            old_value=('foo', 'bar', 'baz'),
            new_value=('foo', 'wombat', 'baz'),
        )
        next_undo_item = UndoItem(
            object=example,
            name='tuple_value',
            old_value=('foo', 'wombat', 'baz'),
            new_value=('foo', 12, 'baz'),
        )

        result = undo_item.merge(next_undo_item)

        self.assertTrue(result)
        self.assertEqual(undo_item.old_value, ('foo', 'bar', 'baz'))
        self.assertEqual(undo_item.new_value, ('foo', 12, 'baz'))

    def test_merge_sequence_change_not_simple_types(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='tuple_value',
            old_value=('foo', ['bar'], 'baz'),
            new_value=('foo', ['wombat'], 'baz'),
        )
        next_undo_item = UndoItem(
            object=example,
            name='tuple_value',
            old_value=('foo', ['wombat'], 'baz'),
            new_value=('foo', ['fizz'], 'baz'),
        )

        result = undo_item.merge(next_undo_item)

        self.assertTrue(result)
        self.assertEqual(undo_item.old_value, ('foo', ['bar'], 'baz'))
        self.assertEqual(undo_item.new_value, ('foo', ['fizz'], 'baz'))

    def test_merge_sequence_change_multiple_not_simple_types(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='tuple_value',
            old_value=(['foo'], 'bar', 'baz'),
            new_value=(['foo'], 'wombat', 'baz'),
        )
        next_undo_item = UndoItem(
            object=example,
            name='tuple_value',
            old_value=(['foo'], 'wombat', 'baz'),
            new_value=(['foo'], 'fizz', 'baz'),
        )

        result = undo_item.merge(next_undo_item)

        # XXX This is perhaps unexpected. See GitHub issue #1507.
        self.assertFalse(result)

    def test_merge_sequence_change_back(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='tuple_value',
            old_value=('foo', 'bar', 'baz'),
            new_value=('foo', 'wombat', 'baz'),
        )
        next_undo_item = UndoItem(
            object=example,
            name='tuple_value',
            old_value=('foo', 'wombat', 'baz'),
            new_value=('foo', 'bar', 'baz'),
        )

        result = undo_item.merge(next_undo_item)

        self.assertFalse(result)

    def test_merge_sequence_two_changes(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='tuple_value',
            old_value=('foo', 'bar', 'baz'),
            new_value=('foo', 'wombat', 'baz'),
        )
        next_undo_item = UndoItem(
            object=example,
            name='tuple_value',
            old_value=('foo', 'wombat', 'baz'),
            new_value=('foo', 'wombat', 'fizz'),
        )

        result = undo_item.merge(next_undo_item)

        self.assertFalse(result)

    def test_merge_sequence_change_length(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='tuple_value',
            old_value=('foo', 'bar', 'baz'),
            new_value=('foo', 'wombat', 'baz'),
        )
        next_undo_item = UndoItem(
            object=example,
            name='tuple_value',
            old_value=('foo', 'wombat', 'baz'),
            new_value=('foo', 'wombat', 'baz', 'fizz'),
        )

        result = undo_item.merge(next_undo_item)

        self.assertFalse(result)

    def test_merge_unhandled_type(self):
        example = SimpleExample()

        undo_item = UndoItem(
            object=example,
            name='any_value',
            old_value={'foo', 'bar', 'baz'},
            new_value={'foo', 'wombat', 'baz'},
        )
        next_undo_item = UndoItem(
            object=example,
            name='any_value',
            old_value={'foo', 'wombat', 'baz'},
            new_value={'foo', 'fizz', 'baz'},
        )

        result = undo_item.merge(next_undo_item)

        self.assertFalse(result)


class TestListUndoItem(UnittestTools, unittest.TestCase):

    def test_undo(self):
        example = SimpleExample(list_value=['foo', 'wombat', 'baz'])

        undo_item = ListUndoItem(
            object=example,
            name='list_value',
            index=1,
            added=['wombat', 'baz'],
            removed=['bar'],
        )

        with self.assertTraitChanges(example, 'list_value_items', count=1):
            undo_item.undo()

        self.assertEqual(example.list_value, ['foo', 'bar'])

    def test_redo(self):
        example = SimpleExample(list_value=['foo', 'bar'])

        undo_item = ListUndoItem(
            object=example,
            name='list_value',
            index=1,
            added=['wombat', 'baz'],
            removed=['bar'],
        )

        with self.assertTraitChanges(example, 'list_value_items', count=1):
            undo_item.redo()

        self.assertEqual(example.list_value, ['foo', 'wombat', 'baz'])

    def test_merge_identical(self):
        example = SimpleExample(list_value=['foo', 'wombat', 'baz'])
        added = [['wombat'], 'baz']
        removed = [['bar']]

        undo_item = ListUndoItem(
            object=example,
            name='list_value',
            index=1,
            added=added.copy(),
            removed=removed.copy(),
        )
        next_undo_item = ListUndoItem(
            object=example,
            name='list_value',
            index=1,
            added=added.copy(),
            removed=removed.copy(),
        )

        result = undo_item.merge(next_undo_item)

        self.assertTrue(result)

    def test_merge_equal(self):
        example = SimpleExample(list_value=['foo', 'wombat', 'baz'])
        removed = [['bar']]

        undo_item = ListUndoItem(
            object=example,
            name='list_value',
            index=1,
            added=[['wombat'], 'baz'],
            removed=removed.copy()
        )
        next_undo_item = ListUndoItem(
            object=example,
            name='list_value',
            index=1,
            added=[['wombat'], 'baz'],
            removed=removed.copy(),
        )

        result = undo_item.merge(next_undo_item)

        self.assertFalse(result)

    def test_merge_equal_removed(self):
        example = SimpleExample(list_value=['foo', 'wombat', 'baz'])
        added = [['wombat'], 'baz']

        undo_item = ListUndoItem(
            object=example,
            name='list_value',
            index=1,
            added=added.copy(),
            removed=[['bar']],
        )
        next_undo_item = ListUndoItem(
            object=example,
            name='list_value',
            index=1,
            added=added.copy(),
            removed=[['bar']],
        )

        result = undo_item.merge(next_undo_item)

        self.assertFalse(result)

    def test_merge_different_objects(self):
        example_1 = SimpleExample()
        example_2 = SimpleExample()
        added = [['wombat'], 'baz']
        removed = [['bar']]

        undo_item = ListUndoItem(
            object=example_1,
            name='list_value',
            index=1,
            added=added.copy(),
            removed=removed.copy(),
        )
        next_undo_item = ListUndoItem(
            object=example_2,
            name='list_value',
            index=1,
            added=added.copy(),
            removed=removed.copy(),
        )

        result = undo_item.merge(next_undo_item)

        self.assertFalse(result)

    def test_merge_different_traits(self):
        example = SimpleExample()
        added = [['wombat'], 'baz']
        removed = [['bar']]

        undo_item = ListUndoItem(
            object=example,
            name='list_value',
            index=1,
            added=added.copy(),
            removed=removed.copy(),
        )
        next_undo_item = ListUndoItem(
            object=example,
            name='any_value',
            index=1,
            added=added.copy(),
            removed=removed.copy(),
        )

        result = undo_item.merge(next_undo_item)

        self.assertFalse(result)

    def test_merge_different_index(self):
        example = SimpleExample()
        added = [['wombat'], 'baz']
        removed = [['bar']]

        undo_item = ListUndoItem(
            object=example,
            name='list_value',
            index=1,
            added=added.copy(),
            removed=removed.copy(),
        )
        next_undo_item = ListUndoItem(
            object=example,
            name='any_value',
            index=0,
            added=added.copy(),
            removed=removed.copy(),
        )

        result = undo_item.merge(next_undo_item)

        self.assertFalse(result)


class TestUndoHistory(UnittestTools, unittest.TestCase):

    def _populate_history(self, history):
        """ Add some simple hostory items. """
        self._example = SimpleExample()
        history.add(
            UndoItem(
                object=self._example,
                name='str_value',
                old_value='foo',
                new_value='bar',
            )
        )
        history.add(
            UndoItem(
                object=self._example,
                name='str_value',
                old_value='bar',
                new_value='wombat',
            ),
        )
        history.add(
            UndoItem(
                object=self._example,
                name='str_value',
                old_value='wombat',
                new_value='baz',
            ),
        )

    def test_defaults(self):
        history = UndoHistory()

        self.assertEqual(history.now, 0)
        self.assertFalse(history.can_undo)
        self.assertFalse(history.can_redo)

    def test_add_empty(self):
        history = UndoHistory()
        example = SimpleExample(str_value='foo')

        undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='foo',
            new_value='bar',
        )

        with self.assertTraitChanges(history, 'undoable', count=1):
            with self.assertTraitDoesNotChange(history, 'redoable'):
                history.add(undo_item)

        self.assertEqual(history.now, 1)
        self.assertTrue(history.can_undo)
        self.assertFalse(history.can_redo)

    def test_add_end(self):
        history = UndoHistory()
        example = SimpleExample(str_value='foo', value=10)
        undo_item = UndoItem(
            object=example,
            name='value',
            old_value=0,
            new_value=10,
        )

        history.add(
            UndoItem(
                object=example,
                name='str_value',
                old_value='foo',
                new_value='bar',
            )
        )

        with self.assertTraitDoesNotChange(history, 'undoable'):
            with self.assertTraitDoesNotChange(history, 'redoable'):
                history.add(undo_item)

        self.assertEqual(history.now, 2)
        self.assertTrue(history.can_undo)
        self.assertFalse(history.can_redo)

    def test_add_merge(self):
        history = UndoHistory()
        example = SimpleExample(str_value='foo', value=10)
        undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='foo',
            new_value='baz',
        )

        history.add(
            UndoItem(
                object=example,
                name='str_value',
                old_value='foo',
                new_value='bar',
            )
        )

        with self.assertTraitDoesNotChange(history, 'undoable'):
            with self.assertTraitDoesNotChange(history, 'redoable'):
                history.add(undo_item)

        self.assertEqual(history.now, 1)
        self.assertTrue(history.can_undo)
        self.assertFalse(history.can_redo)

    def test_add_end_extend(self):
        history = UndoHistory()
        example = SimpleExample(str_value='foo', value=10)
        undo_item = UndoItem(
            object=example,
            name='value',
            old_value=0,
            new_value=10,
        )

        history.add(
            UndoItem(
                object=example,
                name='str_value',
                old_value='foo',
                new_value='bar',
            )
        )

        with self.assertTraitDoesNotChange(history, 'undoable'):
            with self.assertTraitDoesNotChange(history, 'redoable'):
                history.add(undo_item, extend=True)

        self.assertEqual(history.now, 1)
        self.assertTrue(history.can_undo)
        self.assertFalse(history.can_redo)

    def test_add_end_extend_merge(self):
        history = UndoHistory()
        example = SimpleExample(str_value='foo', value=10)
        undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='foo',
            new_value='baz',
        )

        history.add(
            UndoItem(
                object=example,
                name='str_value',
                old_value='foo',
                new_value='bar',
            )
        )

        with self.assertTraitDoesNotChange(history, 'undoable'):
            with self.assertTraitDoesNotChange(history, 'redoable'):
                history.add(undo_item, extend=True)

        self.assertEqual(history.now, 1)
        self.assertTrue(history.can_undo)
        self.assertFalse(history.can_redo)
        # XXX this is testing private state to ensure merge happened
        self.assertEqual(len(history.stack._stack), 1)

    def test_add_middle(self):
        history = UndoHistory()
        example = SimpleExample(str_value='foo', value=10)
        undo_item = UndoItem(
            object=example,
            name='value',
            old_value=0,
            new_value=10,
        )

        history.add(
            UndoItem(
                object=example,
                name='str_value',
                old_value='foo',
                new_value='bar',
            )
        )
        history.add(
            UndoItem(
                object=example,
                name='str_value',
                old_value='bar',
                new_value='wombat',
            ),
        )
        history.undo()

        with self.assertTraitDoesNotChange(history, 'undoable'):
            with self.assertTraitChanges(history, 'redoable', count=1):
                history.add(undo_item)

        self.assertEqual(history.now, 2)
        self.assertTrue(history.can_undo)
        self.assertFalse(history.can_redo)

    def test_add_middle_mergeable(self):
        history = UndoHistory()
        example = SimpleExample(str_value='foo', value=10)
        undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='bar',
            new_value='baz',
        )

        history.add(
            UndoItem(
                object=example,
                name='str_value',
                old_value='foo',
                new_value='bar',
            )
        )
        history.add(
            UndoItem(
                object=example,
                name='str_value',
                old_value='bar',
                new_value='wombat',
            ),
        )
        history.undo()

        with self.assertTraitDoesNotChange(history, 'undoable'):
            with self.assertTraitChanges(history, 'redoable', count=1):
                history.add(undo_item)

        self.assertEqual(history.now, 1)
        self.assertTrue(history.can_undo)
        self.assertFalse(history.can_redo)

    def test_undo_last(self):
        history = UndoHistory()
        self._populate_history(history)

        with self.assertTraitDoesNotChange(history, 'undoable'):
            with self.assertTraitChanges(history, 'redoable', count=1):
                with self.assertTraitChanges(self._example, 'anytrait', count=1):  # noqa: E501
                    history.undo()

        self.assertEqual(history.now, 2)
        self.assertTrue(history.can_undo)
        self.assertTrue(history.can_redo)

    def test_undo_first(self):
        history = UndoHistory()
        self._populate_history(history)
        history.undo()
        history.undo()

        with self.assertTraitDoesNotChange(history, 'redoable'):
            with self.assertTraitChanges(history, 'undoable', count=1):
                with self.assertTraitChanges(self._example, 'anytrait', count=1):  # noqa: E501
                    history.undo()

        self.assertEqual(history.now, 0)
        self.assertFalse(history.can_undo)
        self.assertTrue(history.can_redo)

    def test_undo_middle(self):
        history = UndoHistory()
        self._populate_history(history)
        history.undo()

        with self.assertTraitDoesNotChange(history, 'redoable'):
            with self.assertTraitDoesNotChange(history, 'undoable'):
                with self.assertTraitChanges(self._example, 'anytrait', count=1):  # noqa: E501
                    history.undo()

        self.assertEqual(history.now, 1)
        self.assertTrue(history.can_undo)
        self.assertTrue(history.can_redo)

    def test_redo_last(self):
        history = UndoHistory()
        self._populate_history(history)
        history.undo()

        with self.assertTraitChanges(history, 'redoable', count=1):
            with self.assertTraitDoesNotChange(history, 'undoable'):
                with self.assertTraitChanges(self._example, 'anytrait', count=1):  # noqa: E501
                    history.redo()

        self.assertEqual(history.now, 3)
        self.assertTrue(history.can_undo)
        self.assertFalse(history.can_redo)

    def test_redo_middle(self):
        history = UndoHistory()
        self._populate_history(history)
        history.undo()
        history.undo()

        with self.assertTraitDoesNotChange(history, 'redoable'):
            with self.assertTraitDoesNotChange(history, 'undoable'):
                with self.assertTraitChanges(self._example, 'anytrait', count=1):  # noqa: E501
                    history.redo()

        self.assertEqual(history.now, 2)
        self.assertTrue(history.can_undo)
        self.assertTrue(history.can_redo)

    def test_redo_first(self):
        history = UndoHistory()
        self._populate_history(history)
        history.undo()
        history.undo()
        history.undo()

        with self.assertTraitDoesNotChange(history, 'redoable'):
            with self.assertTraitChanges(history, 'undoable', count=1):
                with self.assertTraitChanges(self._example, 'anytrait', count=1):  # noqa: E501
                    history.redo()

        self.assertEqual(history.now, 1)
        self.assertTrue(history.can_undo)
        self.assertTrue(history.can_redo)

    def test_revert_end(self):
        history = UndoHistory()
        self._populate_history(history)

        with self.assertTraitChanges(history, 'redoable', count=2):
            with self.assertTraitChanges(history, 'undoable', count=1):
                with self.assertTraitChanges(self._example, 'anytrait', count=3):  # noqa: E501
                    history.revert()

        self.assertEqual(history.now, 0)
        self.assertFalse(history.can_undo)
        self.assertFalse(history.can_redo)

    def test_revert_middle(self):
        history = UndoHistory()
        self._populate_history(history)
        history.undo()
        history.undo()

        with self.assertTraitChanges(history, 'redoable', count=1):
            with self.assertTraitChanges(history, 'undoable', count=1):
                with self.assertTraitChanges(self._example, 'anytrait', count=1):  # noqa: E501
                    history.revert()

        self.assertEqual(history.now, 0)
        self.assertFalse(history.can_undo)
        self.assertFalse(history.can_redo)

    def test_revert_start(self):
        history = UndoHistory()
        self._populate_history(history)
        history.undo()
        history.undo()
        history.undo()

        with self.assertTraitDoesNotChange(history, 'undoable'):
            with self.assertTraitChanges(history, 'redoable', count=1):
                with self.assertTraitDoesNotChange(self._example, 'anytrait'):
                    history.revert()

        self.assertEqual(history.now, 0)
        self.assertFalse(history.can_undo)
        self.assertFalse(history.can_redo)

    def test_clear_end(self):
        history = UndoHistory()
        self._populate_history(history)

        with self.assertTraitDoesNotChange(history, 'redoable'):
            with self.assertTraitChanges(history, 'undoable', count=1):
                with self.assertTraitDoesNotChange(self._example, 'anytrait'):
                    history.clear()

        self.assertEqual(history.now, 0)
        self.assertFalse(history.can_undo)
        self.assertFalse(history.can_redo)

    def test_clear_middle(self):
        history = UndoHistory()
        self._populate_history(history)
        history.undo()
        history.undo()

        with self.assertTraitChanges(history, 'redoable', count=1):
            with self.assertTraitChanges(history, 'undoable', count=1):
                with self.assertTraitDoesNotChange(self._example, 'anytrait'):
                    history.clear()

        self.assertEqual(history.now, 0)
        self.assertFalse(history.can_undo)
        self.assertFalse(history.can_redo)

    def test_clear_start(self):
        history = UndoHistory()
        self._populate_history(history)
        history.undo()
        history.undo()
        history.undo()

        with self.assertTraitDoesNotChange(history, 'undoable'):
            with self.assertTraitChanges(history, 'redoable', count=1):
                with self.assertTraitDoesNotChange(self._example, 'anytrait'):
                    history.clear()

        self.assertEqual(history.now, 0)
        self.assertFalse(history.can_undo)
        self.assertFalse(history.can_redo)

    def test_extend(self):
        history = UndoHistory()
        example = SimpleExample(str_value='foo', value=10)
        undo_item = UndoItem(
            object=example,
            name='value',
            old_value=0,
            new_value=10,
        )

        history.add(
            UndoItem(
                object=example,
                name='str_value',
                old_value='foo',
                new_value='bar',
            )
        )

        with self.assertTraitDoesNotChange(history, 'undoable'):
            with self.assertTraitDoesNotChange(history, 'redoable'):
                history.extend(undo_item)

        self.assertEqual(history.now, 1)
        self.assertTrue(history.can_undo)
        self.assertFalse(history.can_redo)

    def test_extend_merge(self):
        history = UndoHistory()
        example = SimpleExample(str_value='foo', value=10)
        undo_item = UndoItem(
            object=example,
            name='str_value',
            old_value='foo',
            new_value='baz',
        )

        history.add(
            UndoItem(
                object=example,
                name='str_value',
                old_value='foo',
                new_value='bar',
            )
        )

        with self.assertTraitDoesNotChange(history, 'undoable'):
            with self.assertTraitDoesNotChange(history, 'redoable'):
                history.extend(undo_item)

        self.assertEqual(history.now, 1)
        self.assertTrue(history.can_undo)
        self.assertFalse(history.can_redo)
        # XXX this is testing private state to ensure merge happened
        self.assertEqual(len(history.stack._stack), 1)

    def test_general_command_do(self):
        history = UndoHistory()
        command = DummyCommand()

        history.add(command)

        self.assertEqual(command.data, "do")


@unittest.skipIf(no_gui_test_assistant, "No GuiTestAssistant")
class TestEditorUndo(BaseTestMixin, GuiTestAssistant, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)
        GuiTestAssistant.setUp(self)

    def tearDown(self):
        GuiTestAssistant.tearDown(self)
        BaseTestMixin.tearDown(self)

    def check_history(self, editor, expected_history_now,
                      expected_history_length):
        # XXX this is testing private state
        return (
            editor.ui.history.now == expected_history_now
            and len(editor.ui.history.stack._stack) == expected_history_length
        )

    def undo(self, editor):
        self.gui.invoke_later(editor.ui.history.undo)
        self.event_loop_helper.event_loop_with_timeout()

    def redo(self, editor):
        self.gui.invoke_later(editor.ui.history.redo)
        self.event_loop_helper.event_loop_with_timeout()

    def test_undo(self):
        editor = create_editor()
        editor.prepare(None)
        editor.ui.history = UndoHistory()

        self.assertEqual(editor.old_value, "test")

        # Enter 'one' followed by 'two'
        with editor.updating_value():
            self.gui.set_trait_later(editor.control, "control_value", "one")
            self.gui.set_trait_later(editor.control, "control_value", "two")

        # Perform an UNDO
        self.undo(editor)

        # Expect 2 items in history and pointer at first item
        self.assertEventuallyTrue(editor, "ui",
                                  functools.partial(self.check_history,
                                                    expected_history_now=1,
                                                    expected_history_length=2),
                                  timeout=5.0)

        # Perform a REDO
        self.redo(editor)

        # Expect 2 items in history and pointer at second item
        self.assertEventuallyTrue(editor, "ui",
                                  functools.partial(self.check_history,
                                                    expected_history_now=2,
                                                    expected_history_length=2),
                                  timeout=5.0)

        # Enter 'three'
        with editor.updating_value():
            self.gui.set_trait_later(editor.control, "control_value", "three")

        # Perform an UNDO
        self.undo(editor)

        # Expect 3 items in history and pointer at second item
        self.assertEventuallyTrue(editor, "ui",
                                  functools.partial(self.check_history,
                                                    expected_history_now=2,
                                                    expected_history_length=3),
                                  timeout=5.0)

        # Enter 'four'
        with editor.updating_value():
            self.gui.set_trait_later(editor.control, "control_value", "four")
        self.event_loop_helper.event_loop_with_timeout()

        # Expect 3 items in history and pointer at second item
        # Note: Modifying the history after an UNDO, clears the future,
        # hence, we expect 3 items in the history, not 4
        self.assertEventuallyTrue(editor, "ui",
                                  functools.partial(self.check_history,
                                                    expected_history_now=3,
                                                    expected_history_length=3),
                                  timeout=5.0)

        # The following sequence after modifying the history had caused
        # the application to hang, verify it.

        # Perform an UNDO
        self.undo(editor)
        self.undo(editor)
        self.redo(editor)

        # Expect 3 items in history and pointer at second item
        self.assertEventuallyTrue(editor, "ui",
                                  functools.partial(self.check_history,
                                                    expected_history_now=2,
                                                    expected_history_length=3),
                                  timeout=5.0)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
import unittest

try:
    import numpy as np  # noqa: F401
except ImportError:
    raise unittest.SkipTest("Can't import NumPy: skipping")
import pkg_resources

from traits.api import Bool, Callable, File, Float, HasTraits, Range, Str
from traitsui.api import ContextValue, Item, View
from traitsui.editors.video_editor import MediaStatus, PlayerState, VideoEditor
from traitsui.tests._tools import BaseTestMixin, create_ui, is_qt5

filename = pkg_resources.resource_filename('traitsui.testing', 'data/test.mp4')


class MovieTheater(HasTraits):
    url = File(filename)

    state = PlayerState()
    duration = Float()
    position = Range(low=0.0, high='duration')
    error = Str()
    status = MediaStatus()
    buffer = Range(0, 100)
    muted = Bool(True)
    volume = Range(0.0, 100.0)
    playback_rate = Float(1.0)
    image_func = Callable()


@unittest.skipIf(not is_qt5(), 'Requires Qt5')
class TestVideoEditor(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def test_video_editor_basics(self):
        obj = MovieTheater()

        # Test an editor with trait syncronizations
        view = View(
            Item(
                'url',
                editor=VideoEditor(
                    state=ContextValue('state'),
                    position=ContextValue('position'),
                    duration=ContextValue('duration'),
                    video_error=ContextValue('error'),
                    media_status=ContextValue('status'),
                    buffer=ContextValue('buffer'),
                    muted=ContextValue('muted'),
                    volume=ContextValue('volume'),
                    playback_rate=ContextValue('playback_rate'),
                    image_func=ContextValue('image_func'),
                    notify_interval=0.5,
                ),
            ),
        )
        with create_ui(obj, {'view': view}):
            pass

        # And an editor with no synced traits
        view = View(Item('url', editor=VideoEditor()))
        with create_ui(obj, {'view': view}):
            pass



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the set editor factory for all traits user interface toolkits.
"""

from traits.api import Bool, Str

from traitsui.editor_factory import EditorWithListFactory


class SetEditor(EditorWithListFactory):
    """ Editor factory for editors for sets.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Are the items in the set ordered (vs. unordered)?
    ordered = Bool(False)

    #: Can the user add and delete all items in the set at once?
    can_move_all = Bool(True)

    #: Title of left column:
    left_column_title = Str()

    #: Title of right column:
    right_column_title = Str()


# This alias is deprecated and will be removed in TraitsUI 8.
ToolkitEditorFactory = SetEditor



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the concrete implementations of the traits Toolkit interface for
    the 'null' (do nothing) user interface toolkit.
"""



from ..toolkit import Toolkit


class GUIToolkit(Toolkit):

    # -------------------------------------------------------------------------
    #  GUI toolkit dependent trait definitions:
    # -------------------------------------------------------------------------

    def color_trait(self, *args, **traits):
        from . import color_trait as ct

        return ct.NullColor(*args, **traits)

    def rgb_color_trait(self, *args, **traits):
        from . import rgb_color_trait as rgbct

        return rgbct.RGBColor(*args, **traits)

    def font_trait(self, *args, **traits):
        from . import font_trait as ft

        return ft.NullFont(*args, **traits)

    def kiva_font_trait(self, *args, **traits):
        from . import font_trait as ft

        return ft.NullFont(*args, **traits)

    def constants(self, *args, **traits):
        constants = {
            "WindowColor": (236 / 255.0, 233 / 255.0, 216 / 255.0, 1.0)
        }
        return constants



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# handler_override.py -- Example of a Handler that overrides setattr(), and
# that has a user interface notification method

from traits.api import Bool, HasTraits
from traitsui.api import Handler, View


class TC_Handler(Handler):

    def setattr(self, info, object, name, value):
        Handler.setattr(self, info, object, name, value)
        info.object._updated = True

    def object__updated_changed(self, info):
        if info.initialized:
            info.ui.title += "*"


class TestClass(HasTraits):
    b1 = Bool()
    b2 = Bool()
    b3 = Bool()
    _updated = Bool(False)


view1 = View(
    'b1',
    'b2',
    'b3',
    title="Alter Title",
    handler=TC_Handler(),
    buttons=['OK', 'Cancel'],
)

tc = TestClass()
tc.configure_traits(view=view1)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
This example demonstrates how to test interacting with a boolean
using BooleanEditor.

The GUI being tested is written in the demo under the same name (minus the
preceding 'test') in the outer directory.
"""

import os
import runpy
import unittest

from traitsui.testing.api import (
    DisplayedText, IsChecked, KeyClick, KeySequence, MouseClick, UITester
)

#: Filename of the demo script
FILENAME = "BooleanEditor_simple_demo.py"

#: Path of the demo script
DEMO_PATH = os.path.join(os.path.dirname(__file__), "..", FILENAME)


class TestBooleanEditorSimpleDemo(unittest.TestCase):

    def test_boolean_editor_simple_demo(self):
        demo = runpy.run_path(DEMO_PATH)["demo"]

        tester = UITester()
        with tester.create_ui(demo) as ui:
            simple = tester.find_by_id(ui, 'simple')
            readonly = tester.find_by_id(ui, 'readonly')
            text = tester.find_by_id(ui, 'text')
            count_changes = tester.find_by_name(ui, "count_changes")

            simple.perform(MouseClick())
            self.assertEqual(demo.my_boolean_trait, True)

            for _ in range(4):
                text.perform(KeyClick("Backspace"))
            text.perform(KeySequence("False"))
            text.perform(KeyClick("Enter"))
            self.assertEqual(demo.my_boolean_trait, False)

            displayed_count_changes = count_changes.inspect(DisplayedText())
            self.assertEqual(displayed_count_changes, '2')
            self.assertEqual(displayed_count_changes, str(demo.count_changes))

            demo.my_boolean_trait = True
            displayed = readonly.inspect(DisplayedText())
            self.assertEqual(displayed, "True")

            simple_is_checked = simple.inspect(IsChecked())
            self.assertEqual(simple_is_checked, True)


# Run the test(s)
unittest.TextTestRunner().run(
    unittest.TestLoader().loadTestsFromTestCase(TestBooleanEditorSimpleDemo)
)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the title editor factory for all traits toolkit backends.
"""

from traits.api import Bool

from traitsui.editor_factory import EditorFactory
from traitsui.toolkit import toolkit_object


class TitleEditor(EditorFactory):
    """ Editor factory for Title editors.
    """

    allow_selection = Bool(False)

    def _get_simple_editor_class(self):
        """ Returns the editor class to use for "simple" style views.
        The default implementation tries to import the SimpleEditor class in
        the editor file in the backend package, and if such a class is not to
        found it returns the SimpleEditor class defined in editor_factory
        module in the backend package.

        """
        SimpleEditor = toolkit_object("title_editor:SimpleEditor")
        return SimpleEditor


# This alias is deprecated and will be removed in TraitsUI 8.
ToolkitEditorFactory = TitleEditor



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Test using a KeyBindings object with the traits CodeEditor."""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from traits.api \
    import HasPrivateTraits, Code, Str

from traitsui.api \
    import View, Item, Handler, CodeEditor

from traitsui.key_bindings \
    import KeyBinding, KeyBindings

#-------------------------------------------------------------------------
#  Define a KeyBindings object:
#-------------------------------------------------------------------------

key_bindings = KeyBindings(

    KeyBinding(binding1='Ctrl-s',
               description='Save to a file',
               method_name='save_file'),

    KeyBinding(binding1='Ctrl-r',
               description='Run script',
               method_name='run_script'),

    KeyBinding(binding1='Ctrl-q',
               description='Edit key bindings',
               method_name='edit_bindings')
)

#-------------------------------------------------------------------------
#  'CodeHandler' class:
#-------------------------------------------------------------------------


class CodeHandler(Handler):

    def save_file(self, info):
        info.object.status = "save file"

    def run_script(self, info):
        info.object.status = "run script"

    def edit_bindings(self, info):
        info.object.status = "edit bindings"
        key_bindings.edit_traits()

#-------------------------------------------------------------------------
#  'TestCode' class:
#-------------------------------------------------------------------------


class TestCode(HasPrivateTraits):

    code = Code()
    status = Str()

    view = View(
        [Item('code',
              style='custom',
              resizable=True,
              editor=CodeEditor(key_bindings=key_bindings)),
         'status~',
         '|<>'],
        id='traitsui.tests.test_code_editor.TestCode',
        title='Sample Code Editor',
        width=0.4,
        height=0.4,
        resizable=True,
        handler=CodeHandler())

#-------------------------------------------------------------------------
#  Run the test:
#-------------------------------------------------------------------------

if __name__ == '__main__':
    TestCode().configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Trait definition for an RGB-based color, which is a tuple of the form
    (*red*, *green*, *blue*), where *red*, *green* and *blue* are floats in the
    range from 0.0 to 1.0.
"""


from traits.api import Trait, TraitError
from traits.trait_base import SequenceTypes

from traitsui.qt4.color_trait import standard_colors

# -------------------------------------------------------------------------
#  Convert a number into an RGB tuple:
# -------------------------------------------------------------------------


def range_check(value):
    """ Checks that *value* can be converted to a value in the range 0.0 to 1.0.

        If so, it returns the floating point value; otherwise, it raises a
        TraitError.
    """
    value = float(value)
    if 0.0 <= value <= 1.0:
        return value
    raise TraitError


def convert_to_color(object, name, value):
    """ Converts a tuple or an integer to an RGB color value, or raises a
        TraitError if that is not possible.
    """
    if isinstance(value, SequenceTypes) and len(value) == 3:
        return (
            range_check(value[0]),
            range_check(value[1]),
            range_check(value[2]),
        )
    if isinstance(value, int):
        return (
            (value / 0x10000) / 255.0,
            ((value // 0x100) & 0xFF) / 255.0,
            (value & 0xFF) / 255.0,
        )
    raise TraitError


convert_to_color.info = (
    "a tuple of the form (r,g,b), where r, g, and b "
    "are floats in the range from 0.0 to 1.0, or an integer which in hex is of "
    "the form 0xRRGGBB, where RR is red, GG is green, and BB is blue"
)

# -------------------------------------------------------------------------
#  Standard colors:
# -------------------------------------------------------------------------

# RGB versions of standard colors:
rgb_standard_colors = {}
for name, color in standard_colors.items():
    rgb_standard_colors[name] = (color.redF(), color.greenF(), color.blueF())

# -------------------------------------------------------------------------
#  Define wxPython specific color traits:
# -------------------------------------------------------------------------

### Note: Declare the editor to be a function which returns the RGBColorEditor
# class from traits ui to avoid circular import issues. For backwards
# compatibility with previous Traits versions, the 'editors' folder in Traits
# project declares 'from api import *' in its __init__.py. The 'api' in turn
# can contain classes that have a RGBColor trait which lead to this file getting
# imported. This will lead to a circular import when declaring a RGBColor
# trait.


def get_rgb_color_editor(*args, **traits):
    from .rgb_color_editor import ToolkitEditorFactory

    return ToolkitEditorFactory(*args, **traits)


# Trait whose value must be an RGB color:
RGBColor = Trait(
    "white", convert_to_color, rgb_standard_colors, editor=get_rgb_color_editor
)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traitsui.testing.tester.locator import TargetById, TargetByName


def _get_editor_by_name(ui, name):
    """ Return a single Editor from an instance of traitsui.ui.UI with
    a given extended name. Raise if zero or many editors are found.

    Parameters
    ----------
    ui : traitsui.ui.UI
        The UI from which an editor will be retrieved.
    name : str
        A single name for retrieving an editor on a UI.

    Returns
    -------
    editor : Editor
        The single editor found.
    """
    editors = ui.get_editors(name)

    all_names = [editor.name for editor in ui._editors]
    if not editors:
        raise ValueError(
            "No editors can be found with name {!r}. "
            "Found these: {!r}".format(name, all_names)
        )
    if len(editors) > 1:
        raise ValueError(
            "Found multiple editors with name {!r}.".format(name))
    editor, = editors
    return editor


def _get_editor_by_id(ui, id):
    """ Return single Editor from an instance of traitsui.ui.UI with
    the given identifier.

    Parameters
    ----------
    ui : traitsui.ui.UI
        The UI from which an editor will be retrieved.
    id : str
        Id for finding an item in the UI.

    Returns
    -------
    editor : Editor
        The single editor found.
    """
    try:
        editor = getattr(ui.info, id)
    except AttributeError:
        raise ValueError(
            "No editors found with id {!r}. Got these: {!r}".format(
                id, ui._names)
            )
    return editor


def register_traitsui_ui_solvers(registry, target_class, traitsui_ui_getter):
    """ Function to register solvers for obtaining nested targets inside a
    traitsui.ui.UI inside a (parent) target.

    For example, an instance of TreeEditor may contain a nested
    ``traitsui.ui.UI`` instance. In that case, the ``target_class`` will the
    TreeEditor and the ``traitsui_ui_getter`` specifies how to obtain the
    nested ``traitsui.ui.UI`` UI from it.

    Parameters
    ----------
    registry : TargetRegistry
        The registry being registered to
    target_class : subclass of type
        The type of a UI target being used as the target_class for the
        solvers
    traitsui_ui_getter : callable(target: target_class) -> traitsui.ui.UI
        A callable specific to the particular target_class to obtain a nested
        UI.
    """

    registry.register_location(
        target_class=target_class,
        locator_class=TargetByName,
        solver=lambda wrapper, location: (
            _get_editor_by_name(
                ui=traitsui_ui_getter(wrapper._target),
                name=location.name,
            )
        ),
    )
    registry.register_location(
        target_class=target_class,
        locator_class=TargetById,
        solver=lambda wrapper, location: (
            _get_editor_by_id(
                ui=traitsui_ui_getter(wrapper._target),
                id=location.id,
            )
        ),
    )



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traits.api import Bool, Constant

from traitsui.editors.date_editor import DateEditor


class DateRangeEditor(DateEditor):
    """ Editor for a date range. The target value should be a tuple
    containing two dates (start date, end date)
    """

    #: This must be set to true for setting a date range.
    multi_select = Constant(True)

    #: Whether it is possible to unset the date range.
    #: If true, then the date range will be set to (None, None)
    #: when all the dates are unselected.
    allow_no_selection = Bool(False)


# This alias is deprecated and will be removed in TraitsUI 8.
ToolkitEditorFactory = DateRangeEditor



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the tree-based Python value editor and the value editor factory,
    for the wxPython user interface toolkit.
"""


from traitsui.editors.value_editor import _ValueEditor

from .editor import Editor


class SimpleEditor(_ValueEditor, Editor):
    """ Returns the editor to use for simple style views.
    """

    #: Override the value of the readonly trait.
    readonly = False


class ReadonlyEditor(_ValueEditor, Editor):
    """ Returns the editor to use for readonly style views.
    """

    #: Override the value of the readonly trait.
    readonly = True



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.api import HasTraits, Int
from traitsui.api import Item, View
from traitsui.tests._tools import (
    create_ui,
    requires_toolkit,
    ToolkitName,
)


class ObjectWithNumber(HasTraits):
    number = Int()


@requires_toolkit([ToolkitName.qt])
class TestStickyDialog(unittest.TestCase):
    """ Test _StickyDialog used by the UI's Qt backend.
    """

    def test_sticky_dialog_with_parent(self):
        obj = ObjectWithNumber()
        obj2 = ObjectWithNumber()
        parent_view = View(Item("number"), title="Parent")
        nested = View(
            Item("number"),
            resizable=True,
            title="Nested"
        )
        with create_ui(obj, dict(view=parent_view)) as ui:
            with create_ui(obj2, dict(parent=ui.control, view=nested)) as ui2:
                from pyface.qt import QtCore
                self.assertFalse(
                    ui2.control.windowFlags() & QtCore.Qt.WindowMaximized
                )



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A single line text widget that supports functionality common to native
    search widgets.
"""

from traits.api import Bool, Property, Str

from traitsui.basic_editor_factory import BasicEditorFactory
from traitsui.toolkit import toolkit_object


class SearchEditor(BasicEditorFactory):
    """ A single line text widget that supports functionality common to native
        search widgets.
    """

    #: The editor class to be created:
    klass = Property()

    #: The descriptive text for the widget
    text = Str("Search")

    #: Is user input set on every keystroke?
    auto_set = Bool(True)

    #: Is user input set when the Enter key is pressed?
    enter_set = Bool(False)

    #: Whether to show a search button on the widget
    search_button = Bool(True)

    #: Whether to show a cancel button on the widget
    cancel_button = Bool(False)

    #: Fire this event on the object whenever a search should be triggered,
    #: regardless of whether the search term changed
    search_event_trait = Str()

    def _get_klass(self):
        """ Returns the toolkit-specific editor class to be instantiated.
        """
        return toolkit_object("search_editor:SearchEditor")



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the various editors for single-selection enumerations, for the
wxPython user interface toolkit.
"""


import wx

from traits.api import Property

from traitsui.helper import enum_values_changed

from .editor import Editor

from .constants import OKColor, ErrorColor

from .helper import (
    TraitsUIPanel,
    disconnect,
    disconnect_no_id,
)
from functools import reduce

# -------------------------------------------------------------------------
#  'BaseEditor' class:
# -------------------------------------------------------------------------


class BaseEditor(Editor):
    """ Base class for enumeration editors.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Current set of enumeration names:
    names = Property()

    #: Current mapping from names to values:
    mapping = Property()

    #: Current inverse mapping from values to names:
    inverse_mapping = Property()

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        factory = self.factory
        if factory.name != "":
            self._object, self._name, self._value = self.parse_extended_name(
                factory.name
            )
            self.values_changed()
            self._object.on_trait_change(
                self._values_changed, " " + self._name, dispatch="ui"
            )
        else:
            self._value = lambda: self.factory.values
            self.values_changed()
            factory.on_trait_change(
                self._values_changed, "values", dispatch="ui"
            )

    def _get_names(self):
        """ Gets the current set of enumeration names.
        """
        return self._names

    def _get_mapping(self):
        """ Gets the current mapping.
        """
        return self._mapping

    def _get_inverse_mapping(self):
        """ Gets the current inverse mapping.
        """
        return self._inverse_mapping

    def rebuild_editor(self):
        """ Rebuilds the contents of the editor whenever the original factory
            object's **values** trait changes.
        """
        raise NotImplementedError

    def values_changed(self):
        """ Recomputes the cached data based on the underlying enumeration model
            or the values of the factory.
        """
        self._names, self._mapping, self._inverse_mapping = enum_values_changed(
            self._value(), self.string_value
        )

    def _values_changed(self):
        """ Handles the underlying object model's enumeration set or factory's
            values being changed.
        """
        self.values_changed()
        self.rebuild_editor()

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        if self._object is not None:
            self._object.on_trait_change(
                self._values_changed, " " + self._name, remove=True
            )
        else:
            self.factory.on_trait_change(
                self._values_changed, "values", remove=True
            )

        super().dispose()


# -------------------------------------------------------------------------
#  'SimpleEditor' class:
# -------------------------------------------------------------------------


class SimpleEditor(BaseEditor):
    """ Simple style of enumeration editor, which displays a combo box.
    """

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        super().init(parent)

        factory = self.factory

        if factory.evaluate is None:
            self.control = control = wx.Choice(
                parent, -1, wx.Point(0, 0), wx.Size(-1, -1), self.names
            )
            parent.Bind(wx.EVT_CHOICE, self.update_object, id=self.control.GetId())
        else:
            self.control = control = wx.ComboBox(
                parent,
                -1,
                "",
                wx.Point(0, 0),
                wx.Size(-1, -1),
                self.names,
                style=wx.CB_DROPDOWN,
            )
            parent.Bind(wx.EVT_COMBOBOX, self.update_object, id=control.GetId())
            parent.Bind(wx.EVT_TEXT_ENTER, self.update_text_object, id=control.GetId())
            control.Bind(wx.EVT_KILL_FOCUS, self.on_kill_focus)

            if (not factory.is_grid_cell) and factory.auto_set:
                parent.Bind(wx.EVT_TEXT, self.update_text_object, id=control.GetId())

        self._no_enum_update = 0
        self.set_tooltip()

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        disconnect(
            self.control, wx.EVT_COMBOBOX, wx.EVT_TEXT_ENTER, wx.EVT_TEXT
        )

        disconnect_no_id(self.control, wx.EVT_KILL_FOCUS)

        super().dispose()

    def update_object(self, event):
        """ Handles the user selecting a new value from the combo box.
        """
        self._no_enum_update += 1
        try:
            new_value = self.mapping[event.GetString()]
            if new_value == self.value and self.factory.is_grid_cell:
                # If the enum editor is in a grid cell and the value did not
                # change, we want the enum editor to go away, reverting back to
                # the normal cell appearance. This is for 2 reasons:
                #  1. it looks nicer
                #  2. if the grid id suddenly closed, wx freaks & causes a
                #     segfault

                grid = self.control.Parent.Parent
                grid.EnableEditing(False)
                grid.EnableEditing(True)

            self.value = new_value

        except:
            from traitsui.api import raise_to_debug

            raise_to_debug()

        self._no_enum_update -= 1

    def update_text_object(self, event):
        """ Handles the user typing text into the combo box text entry field.
        """
        if self._no_enum_update == 0:
            value = self.control.GetValue()
            try:
                value = self.mapping[value]
            except:
                try:
                    value = self.factory.evaluate(value)
                except Exception as excp:
                    self.error(excp)
                    return

            self._no_enum_update += 1
            try:
                self.value = value
                self.control.SetBackgroundColour(OKColor)
                self.control.Refresh()
            except:
                pass
            self._no_enum_update -= 1

    def on_kill_focus(self, event):
        """ Handles the control losing the keyboard focus.
        """
        self.update_text_object(event)
        event.Skip()

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        if self._no_enum_update == 0:
            if self.factory.evaluate is None:
                try:
                    self.control.SetStringSelection(
                        self.inverse_mapping[self.value]
                    )
                except:
                    pass
            else:
                try:
                    self.control.SetValue(self.str_value)
                except:
                    pass

    def error(self, excp):
        """ Handles an error that occurs while setting the object's trait value.
        """
        self.control.SetBackgroundColour(ErrorColor)
        self.control.Refresh()

    def rebuild_editor(self):
        """ Rebuilds the contents of the editor whenever the original factory
            object's **values** trait changes.
        """
        # Note: This code is unnecessarily complex due to a strange bug in
        # wxWidgets implementation of the wx.Combobox control that has strange
        # behavior when the current text field value is one of the selection
        # values when 'Clear' is called. In this case, even saving and
        # restoring the text field value does not work, so we go to great
        # lengths to detect this case and avoid using 'Clear', but still get
        # the equivalent visual results. Modify this code at your own risk...

        control = self.control
        clear = True
        cur_name = None
        if self.factory.evaluate is not None:
            n = control.GetCount()
            cur_names = [control.GetString(i) for i in range(n)]
            cur_name = control.GetValue()
            if cur_name in self.names:
                clear = False
                include = True
                for i in range(n - 1, -1, -1):
                    if cur_name == cur_names[i]:
                        include = False
                    else:
                        control.Delete(i)
                for name in self.names:
                    if include or (name != cur_name):
                        control.Append(name)
                cur_name = None
            else:
                point = control.GetInsertionPoint()

        if clear:
            control.Clear()
            control.AppendItems(self.names)

        if cur_name is not None:
            self._no_enum_update += 1
            control.SetValue(cur_name)
            control.SetInsertionPoint(point)
            self._no_enum_update -= 1

        self.update_editor()


class RadioEditor(BaseEditor):
    """ Enumeration editor, used for the "custom" style, that displays radio
        buttons.
    """

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        super().init(parent)

        # Create a panel to hold all of the radio buttons:
        self.control = TraitsUIPanel(parent, -1)
        self.rebuild_editor()

    def update_object(self, event):
        """ Handles the user clicking one of the custom radio buttons.
        """
        try:
            self.value = event.GetEventObject().value
        except:
            pass

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        value = self.value
        for button in self.control.GetChildren():
            state = button.value == value
            button.SetValue(state)
            if state:
                button.SetFocus()

    def rebuild_editor(self):
        """ Rebuilds the contents of the editor whenever the original factory
            object's **values** trait changes.
        """
        # Clear any existing content:
        panel = self.control
        panel.SetSizer(None)
        panel.DestroyChildren()

        # Get the current trait value:
        cur_name = self.str_value

        # Create a sizer to manage the radio buttons:
        names = self.names
        mapping = self.mapping
        n = len(names)
        cols = self.factory.cols
        rows = (n + cols - 1) // cols
        # incr will keep track of how to increment index so that as we traverse
        # the grid in row major order, the elements are added to appear in
        # column major order
        incr = [n // cols] * cols
        rem = n % cols
        for i in range(cols):
            incr[i] += rem > i
        incr[-1] = -(reduce(lambda x, y: x + y, incr[:-1], 0) - 1)
        # e.g for a gird:
        # 0 2 4
        # 1 3 5
        # incr should be [2, 2, -3]
        if cols > 1:
            sizer = wx.GridSizer(0, cols, 2, 4)
        else:
            sizer = wx.BoxSizer(wx.VERTICAL)

        # Add the set of all possible choices:
        style = wx.RB_GROUP
        index = 0
        # populate the layout in row_major order
        for i in range(rows):
            for j in range(cols):
                if n > 0:
                    name = label = names[index]
                    label = self.string_value(label, str.capitalize)
                    control = wx.RadioButton(panel, -1, label, style=style)
                    control.value = mapping[name]
                    style = 0
                    control.SetValue(name == cur_name)
                    panel.Bind(wx.EVT_RADIOBUTTON, self.update_object, id= control.GetId())
                    self.set_tooltip(control)
                    index += incr[j]
                    n -= 1
                else:
                    control = wx.RadioButton(panel, -1, "")
                    control.value = ""
                    control.Show(False)
                sizer.Add(control, 0, wx.NORTH, 5)

        # Set-up the layout:
        panel.SetSizerAndFit(sizer)


class ListEditor(BaseEditor):
    """ Enumeration editor, used for the "custom" style, that displays a list
        box.
    """

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        super().init(parent)

        # Create a panel to hold all of the radio buttons:
        self.control = wx.ListBox(
            parent,
            -1,
            wx.Point(0, 0),
            wx.Size(-1, -1),
            self.names,
            style=wx.LB_SINGLE | wx.LB_NEEDED_SB,
        )
        parent.Bind(wx.EVT_LISTBOX, self.update_object, id=self.control.GetId())
        self.set_tooltip()

    def dispose(self):
        """ Disposes of the contents of an editor.
        """
        disconnect(self.control, wx.EVT_LISTBOX)

        super().dispose()

    def update_object(self, event):
        """ Handles the user selecting a list box item.
        """
        if not self._ignore_update:
            value = self.control.GetStringSelection()
            try:
                value = self.mapping[value]
            except:
                try:
                    value = self.factory.evaluate(value)
                except:
                    pass
            try:
                self.value = value
            except:
                pass

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        control = self.control
        try:
            index = control.FindString(self.inverse_mapping[self.value])
            if index >= 0:
                control.SetSelection(index)
        except:
            pass

    def rebuild_editor(self):
        """ Rebuilds the contents of the editor whenever the original factory
            object's **values** trait changes.
        """
        self._ignore_update = True
        self.control.Clear()
        self.control.AppendItems(self.names)
        self._ignore_update = False

        # fixme: Is this line necessary?
        self.update_editor()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import platform
import unittest

from traits.api import HasTraits, Float, Int
from traitsui.api import Item, RangeEditor, UItem, View
from traitsui.testing.api import (
    DisplayedText,
    KeyClick,
    KeySequence,
    Slider,
    TargetRegistry,
    Textbox,
    UITester
)
from traitsui.tests._tools import (
    BaseTestMixin,
    is_wx,
    requires_toolkit,
    ToolkitName,
)

is_windows = platform.system() == "Windows"


def _register_simple_spin(registry):
    """ Register interactions for the given registry for a SimpleSpinEditor.

    If there are any conflicts, an error will occur.

    This is kept separate from the below register function because the
    SimpleSpinEditor is not yet implemented on wx.  This function can be used
    with a local reigstry for tests.

    Parameters
    ----------
    registry : TargetRegistry
        The registry being registered to.
    """
    from traitsui.testing.tester._ui_tester_registry.qt4 import (
        _registry_helper
    )
    from traitsui.qt4.range_editor import SimpleSpinEditor

    _registry_helper.register_editable_textbox_handlers(
        registry=registry,
        target_class=SimpleSpinEditor,
        widget_getter=lambda wrapper: wrapper._target.control.lineEdit(),
    )


class RangeModel(HasTraits):

    value = Int(1)
    float_value = Float(0.1)


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestRangeEditor(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def check_range_enum_editor_format_func(self, style):
        # RangeEditor with enum mode doesn't support format_func
        obj = RangeModel()
        view = View(
            UItem(
                "value",
                editor=RangeEditor(
                    low=1, high=3,
                    format_func=lambda v: "{:02d}".format(v),
                    mode="enum"
                ),
                style=style,
            )
        )

        tester = UITester()
        with tester.create_ui(obj, dict(view=view)) as ui:
            editor = ui.get_editors("value")[0]

            # No formatting - simple strings
            self.assertEqual(editor.names[:3], ["1", "2", "3"])
            self.assertEqual(editor.mapping, {"1": 1, "2": 2, "3": 3})
            self.assertEqual(
                editor.inverse_mapping, {1: "1", 2: "2", 3: "3"}
            )

    def test_simple_editor_format_func(self):
        self.check_range_enum_editor_format_func("simple")

    def test_custom_editor_format_func(self):
        self.check_range_enum_editor_format_func("custom")

    def check_slider_set_with_text_valid(self, mode):
        model = RangeModel()
        view = View(
            Item(
                "value",
                editor=RangeEditor(low=1, high=12, mode=mode)
            )
        )
        tester = UITester()
        with tester.create_ui(model, dict(view=view)) as ui:
            # sanity check
            self.assertEqual(model.value, 1)
            number_field = tester.find_by_name(ui, "value")
            text = number_field.locate(Textbox())
            text.perform(KeyClick("0"))
            text.perform(KeyClick("Enter"))
            displayed = text.inspect(DisplayedText())
            self.assertEqual(model.value, 10)
            self.assertEqual(displayed, str(model.value))

    def test_simple_slider_editor_set_with_text_valid(self):
        return self.check_slider_set_with_text_valid(mode='slider')

    def test_large_range_slider_editor_set_with_text_valid(self):
        return self.check_slider_set_with_text_valid(mode='xslider')

    def test_log_range_slider_editor_set_with_text_valid(self):
        return self.check_slider_set_with_text_valid(mode='logslider')

    def test_range_text_editor_set_with_text_valid(self):
        model = RangeModel()
        view = View(
            Item(
                "value",
                editor=RangeEditor(low=1, high=12, mode="text")
            )
        )
        tester = UITester()
        with tester.create_ui(model, dict(view=view)) as ui:
            # sanity check
            self.assertEqual(model.value, 1)
            number_field_text = tester.find_by_name(ui, "value")
            if is_windows and is_wx():
                # For RangeTextEditor on wx and windows, the textbox
                # automatically gets focus and the full content is selected.
                # Insertion point is moved to keep the test consistent
                number_field_text.perform(KeyClick("End"))
            number_field_text.perform(KeyClick("0"))
            number_field_text.perform(KeyClick("Enter"))
            displayed = number_field_text.inspect(DisplayedText())
            self.assertEqual(model.value, 10)
            self.assertEqual(displayed, str(model.value))

    # the tester support code is not yet implemented for Wx SimpleSpinEditor
    @requires_toolkit([ToolkitName.qt])
    def test_simple_spin_editor_set_with_text_valid(self):
        model = RangeModel()
        view = View(
            Item(
                "value",
                editor=RangeEditor(low=1, high=12, mode="spinner")
            )
        )
        LOCAL_REGISTRY = TargetRegistry()
        _register_simple_spin(LOCAL_REGISTRY)
        tester = UITester(registries=[LOCAL_REGISTRY])
        with tester.create_ui(model, dict(view=view)) as ui:
            # sanity check
            self.assertEqual(model.value, 1)
            number_field = tester.find_by_name(ui, "value")
            # For whatever reason, "End" was not working here
            number_field.perform(KeyClick("Right"))
            number_field.perform(KeyClick("0"))
            number_field.perform(KeyClick("Enter"))
            displayed = number_field.inspect(DisplayedText())
            self.assertEqual(model.value, 10)
            self.assertEqual(displayed, str(model.value))

    def check_slider_set_with_text_after_empty(self, mode):
        model = RangeModel()
        view = View(
            Item(
                "value",
                editor=RangeEditor(low=1, high=12, mode=mode)
            )
        )
        tester = UITester()
        with tester.create_ui(model, dict(view=view)) as ui:
            number_field = tester.find_by_name(ui, "value")
            text = number_field.locate(Textbox())
            # Delete all contents of textbox
            for _ in range(5):
                text.perform(KeyClick("Backspace"))
            text.perform(KeySequence("11"))
            text.perform(KeyClick("Enter"))
            displayed = text.inspect(DisplayedText())
            self.assertEqual(model.value, 11)
            self.assertEqual(displayed, str(model.value))

    def test_simple_slider_editor_set_with_text_after_empty(self):
        return self.check_slider_set_with_text_after_empty(mode='slider')

    def test_large_range_slider_editor_set_with_text_after_empty(self):
        return self.check_slider_set_with_text_after_empty(mode='xslider')

    def test_log_range_slider_editor_set_with_text_after_empty(self):
        return self.check_slider_set_with_text_after_empty(mode='logslider')

    # on wx the text style editor gives an error whenever the textbox
    # is empty, even if enter has not been pressed.
    @requires_toolkit([ToolkitName.qt])
    def test_range_text_editor_set_with_text_after_empty(self):
        model = RangeModel()
        view = View(
            Item(
                "value",
                editor=RangeEditor(low=1, high=12, mode="text")
            )
        )
        tester = UITester()
        with tester.create_ui(model, dict(view=view)) as ui:
            number_field_text = tester.find_by_name(ui, "value")
            # Delete all contents of textbox
            for _ in range(5):
                number_field_text.perform(KeyClick("Backspace"))
            number_field_text.perform(KeySequence("11"))
            number_field_text.perform(KeyClick("Enter"))
            displayed = number_field_text.inspect(DisplayedText())
            self.assertEqual(model.value, 11)
            self.assertEqual(displayed, str(model.value))

    # the tester support code is not yet implemented for Wx SimpleSpinEditor
    @requires_toolkit([ToolkitName.qt])
    def test_simple_spin_editor_set_with_text_after_empty(self):
        model = RangeModel()
        view = View(
            Item(
                "value",
                editor=RangeEditor(low=1, high=12, mode="spinner")
            )
        )
        LOCAL_REGISTRY = TargetRegistry()
        _register_simple_spin(LOCAL_REGISTRY)
        tester = UITester(registries=[LOCAL_REGISTRY])
        with tester.create_ui(model, dict(view=view)) as ui:
            number_field_text = tester.find_by_name(ui, "value")
            number_field_text.perform(KeyClick("Right"))
            # Delete all contents of textbox
            for _ in range(5):
                number_field_text.perform(KeyClick("Backspace"))
            number_field_text.perform(KeySequence("11"))
            number_field_text.perform(KeyClick("Enter"))
            displayed = number_field_text.inspect(DisplayedText())
            self.assertEqual(model.value, 11)
            self.assertEqual(displayed, str(model.value))

    def check_modify_slider(self, mode):
        model = RangeModel(value=0)
        view = View(
            Item(
                "value",
                editor=RangeEditor(low=0, high=10, mode=mode)
            )
        )
        tester = UITester()
        with tester.create_ui(model, dict(view=view)) as ui:
            number_field = tester.find_by_name(ui, "value")
            slider = number_field.locate(Slider())
            text = number_field.locate(Textbox())
            # slider values are converted to a [0, 10000] scale.  A single
            # step is a change of 100 on that scale and a page step is 1000.
            # Our range in [0, 10] so these correspond to changes of .1 and 1.
            # Note: when tested manually, the step size seen on OSX and Wx is
            # different.
            for _ in range(10):
                slider.perform(KeyClick("Right"))
            displayed = text.inspect(DisplayedText())
            self.assertEqual(model.value, 1)
            self.assertEqual(displayed, str(model.value))
            slider.perform(KeyClick("Page Up"))
            displayed = text.inspect(DisplayedText())
            self.assertEqual(model.value, 2)
            self.assertEqual(displayed, str(model.value))

    def test_modify_slider_simple_slider(self):
        return self.check_modify_slider('slider')

    def test_modify_slider_large_range_slider(self):
        return self.check_modify_slider('xslider')

    def test_modify_slider_log_range_slider(self):
        model = RangeModel()
        view = View(
            Item(
                "float_value",
                editor=RangeEditor(low=.1, high=1000000000, mode='logslider')
            )
        )
        tester = UITester()
        with tester.create_ui(model, dict(view=view)) as ui:
            number_field = tester.find_by_name(ui, "float_value")
            slider = number_field.locate(Slider())
            text = number_field.locate(Textbox())
            # 10 steps is equivalent to 1 page step
            # on this scale either of those is equivalent to increasing the
            # trait's value from 10^n to 10^(n+1)
            for _ in range(10):
                slider.perform(KeyClick("Right"))
            displayed = text.inspect(DisplayedText())
            self.assertEqual(model.float_value, 1.0)
            self.assertEqual(displayed, str(model.float_value))
            slider.perform(KeyClick("Page Up"))
            displayed = text.inspect(DisplayedText())
            self.assertEqual(model.float_value, 10.0)
            self.assertEqual(displayed, str(model.float_value))



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.api import HasTraits, Bool
from traitsui.api import BooleanEditor, Item, View
from traitsui.tests._tools import (
    BaseTestMixin,
    requires_toolkit,
    ToolkitName,
)

from traitsui.testing.api import (
    DisplayedText, IsChecked, KeyClick, KeySequence, MouseClick, UITester
)


class BoolModel(HasTraits):

    true_or_false = Bool()


# Run this against wx once enthought/traitsui#752 is also fixed for
# BooleanEditor
@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestBooleanEditor(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def check_click_boolean_changes_trait(self, style):
        view = View(Item("true_or_false", style=style, editor=BooleanEditor()))
        obj = BoolModel()

        tester = UITester()
        with tester.create_ui(obj, dict(view=view)) as ui:
            # sanity check
            self.assertEqual(obj.true_or_false, False)
            checkbox = tester.find_by_name(ui, "true_or_false")
            checkbox.perform(MouseClick())
            self.assertEqual(obj.true_or_false, True)

    def test_click_boolean_changes_trait_simple(self):
        self.check_click_boolean_changes_trait('simple')

    def test_click_boolean_changes_trait_custom(self):
        self.check_click_boolean_changes_trait('custom')

    def test_change_text_boolean_changes_trait(self):
        view = View(Item("true_or_false", style='text'))
        obj = BoolModel()

        tester = UITester()
        with tester.create_ui(obj, dict(view=view)) as ui:
            self.assertEqual(obj.true_or_false, False)
            text_field = tester.find_by_name(ui, 'true_or_false')
            for _ in range(5):
                text_field.perform(KeyClick("Backspace"))
            text_field.perform(KeySequence("True"))
            text_field.perform(KeyClick("Enter"))
            self.assertEqual(obj.true_or_false, True)

    def check_trait_change_shown_in_gui(self, style):
        view = View(Item("true_or_false", style=style))
        obj = BoolModel()

        tester = UITester()
        with tester.create_ui(obj, dict(view=view)) as ui:
            checkbox = tester.find_by_name(ui, "true_or_false")
            checked = checkbox.inspect(IsChecked())
            # sanity check
            self.assertEqual(checked, False)
            obj.true_or_false = True
            checked = checkbox.inspect(IsChecked())
            self.assertEqual(checked, True)

    def test_trait_change_shown_in_gui_simple(self):
        self.check_trait_change_shown_in_gui('simple')

    def test_trait_change_shown_in_gui_custom(self):
        self.check_trait_change_shown_in_gui('custom')

    def test_trait_change_shown_in_gui_readonly(self):
        view = View(Item("true_or_false", style='readonly'))
        obj = BoolModel()

        tester = UITester()
        with tester.create_ui(obj, dict(view=view)) as ui:
            readonly = tester.find_by_name(ui, "true_or_false")
            displayed = readonly.inspect(DisplayedText())
            # sanity check
            self.assertEqual(displayed, 'False')
            obj.true_or_false = True
            displayed = readonly.inspect(DisplayedText())
            self.assertEqual(displayed, 'True')

    def test_trait_change_shown_in_gui_text(self):
        view = View(Item("true_or_false", style='text'))
        obj = BoolModel()

        tester = UITester()
        with tester.create_ui(obj, dict(view=view)) as ui:
            text_field = tester.find_by_name(ui, "true_or_false")
            displayed = text_field.inspect(DisplayedText())
            # sanity check
            self.assertEqual(displayed, 'False')
            obj.true_or_false = True
            displayed = text_field.inspect(DisplayedText())
            self.assertEqual(displayed, 'True')



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" This module defines interaction objects that can be passed to
``UIWrapper.perform`` where the actions represent 'commands'.

Implementations for these actions are expected to produce the
documented side effects without returning any values.
"""


class MouseClick:
    """ An object representing the user clicking a mouse button.
    Currently the left mouse button is assumed.

    In most circumstances, a widget can still be clicked on even if it is
    disabled. Therefore unlike key events, if the widget is disabled,
    implementations should not raise an exception.
    """
    pass


class KeySequence:
    """ An object representing the user typing a sequence of keys.

    Implementations should raise ``Disabled`` if the widget is disabled.

    Attributes
    ----------
    sequence : str
        A string that represents a sequence of key inputs.
        e.g. "Hello World"
    """

    def __init__(self, sequence):
        self.sequence = sequence


class KeyClick:
    """ An object representing the user clicking a key on the keyboard.

    Implementations should raise ``Disabled`` if the widget is disabled.

    Attributes
    ----------
    key : str
        Standardized (pyface) name for a keyboard event.
        e.g. "Enter", "Tab", "Space", "0", "1", "A", ...
    """

    def __init__(self, key):
        self.key = key



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
import shutil
import tempfile
import unittest

from pyface.toolkit import toolkit_object
from traits.api import (
    Directory, HasStrictTraits, Instance, Int, List, Str, TraitError
)

from traitsui.api import Item, ListEditor, View
from traitsui.testing.api import (
    DisplayedText,
    Index,
    KeyClick,
    KeySequence,
    LocationNotSupported,
    MouseClick,
    Textbox,
    UITester
)
from traitsui.tests._tools import (
    requires_toolkit,
    ToolkitName,
)

ModalDialogTester = toolkit_object(
    "util.modal_dialog_tester:ModalDialogTester"
)


# 'Person' class:
class Person(HasStrictTraits):

    # Trait definitions:
    name = Str()
    age = Int()

    # Traits view definition:
    traits_view = View(
        'name', 'age',
        width=0.18,
        buttons=['OK', 'Cancel']
    )


def get_people():
    # Sample data:
    return [
        Person(name='Dave', age=39),
        Person(name='Mike', age=28),
        Person(name='Joe', age=34),
        Person(name='Tom', age=22),
        Person(name='Dick', age=63),
        Person(name='Harry', age=46),
        Person(name='Sally', age=43),
        Person(name='Fields', age=31)
    ]


# 'ListTraitTest' class:
class ListTraitTest(HasStrictTraits):

    # Trait definitions:
    people = List(Instance(Person, ()))
    num_columns = Int(1)
    style = Str("custom")

    # Traits view definitions:
    def default_traits_view(self):
        view = View(
            Item(
                'people',
                label='List',
                id='list',
                style=self.style,
                editor=ListEditor(style=self.style, columns=self.num_columns),
            ),
            resizable=True
        )
        return view


class Phonebook(HasStrictTraits):
    people = List(Instance(Person))


notebook_view = View(
    Item(
        "people",
        style="custom",
        editor=ListEditor(use_notebook=True),
    )
)


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestCustomListEditor(unittest.TestCase):

    def test_locate_element_and_edit(self):
        # varying the number of columns in the view tests the logic for
        # getting the correct nested ui
        for col in range(1, 5):
            obj = ListTraitTest(people=get_people(), num_columns=col)
            tester = UITester()
            with tester.create_ui(obj) as ui:
                # sanity check
                self.assertEqual(obj.people[7].name, "Fields")
                people_list = tester.find_by_name(ui, "people")
                item = people_list.locate(Index(7))
                name_field = item.find_by_name("name")
                for _ in range(6):
                    name_field.perform(KeyClick("Backspace"))
                name_field.perform(KeySequence("David"))
                displayed = name_field.inspect(DisplayedText())
                self.assertEqual(obj.people[7].name, "David")
                self.assertEqual(displayed, obj.people[7].name)

    def test_useful_err_message(self):
        obj = ListTraitTest(people=get_people())
        tester = UITester()
        with tester.create_ui(obj) as ui:
            with self.assertRaises(LocationNotSupported) as exc:
                people_list = tester.find_by_name(ui, "people")
                people_list.locate(Textbox())
            self.assertIn(Index, exc.exception.supported)

    def test_index_out_of_range(self):
        obj = ListTraitTest(people=get_people())
        tester = UITester()
        with tester.create_ui(obj) as ui:
            people_list = tester.find_by_name(ui, "people")
            with self.assertRaises(IndexError):
                people_list.locate(Index(10))


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestSimpleListEditor(unittest.TestCase):

    def test_locate_element_and_edit(self):
        obj = ListTraitTest(people=get_people(), style="simple")
        tester = UITester()
        with tester.create_ui(obj) as ui:
            # sanity check
            self.assertEqual(obj.people[7].name, "Fields")
            people_list = tester.find_by_name(ui, "people")
            item = people_list.locate(Index(7))
            item.perform(MouseClick())
            name_field = item.find_by_name("name")
            for _ in range(6):
                name_field.perform(KeyClick("Backspace"))
            name_field.perform(KeySequence("David"))
            displayed = name_field.inspect(DisplayedText())
            self.assertEqual(obj.people[7].name, "David")
            self.assertEqual(displayed, obj.people[7].name)

    def test_useful_err_message(self):
        obj = ListTraitTest(people=get_people(), style="simple")
        tester = UITester()
        with tester.create_ui(obj) as ui:
            with self.assertRaises(LocationNotSupported) as exc:
                people_list = tester.find_by_name(ui, "people")
                people_list.locate(Textbox())
            self.assertIn(Index, exc.exception.supported)

    def test_index_out_of_range(self):
        obj = ListTraitTest(people=get_people(), style="simple")
        tester = UITester()
        with tester.create_ui(obj) as ui:
            people_list = tester.find_by_name(ui, "people")
            with self.assertRaises(IndexError):
                people_list.locate(Index(10))

    # regression test for enthought/traitsui#1154
    @requires_toolkit([ToolkitName.qt])
    def test_add_item_fails(self):

        class Foo(HasStrictTraits):
            dirs = List(Directory(exists=True))

        obj = Foo()
        tester = UITester(auto_process_events=False)
        with tester.create_ui(obj) as ui:
            dirs_list_editor = tester.find_by_name(ui, "dirs")

            def trigger_error():
                try:
                    dirs_list_editor._target.add_empty()
                except TraitError:
                    return False
                return True

            mdtester = ModalDialogTester(trigger_error)
            mdtester.open_and_run(lambda x: x.close())
            # we want an error dialog to open, but don't want to raise a
            # TraitError and crash the full application
            self.assertTrue(mdtester.dialog_was_opened)
            self.assertTrue(mdtester.result)

            self.assertEqual(len(obj.dirs), 0)

    # this test hits a problem on wx, see issue enthought/traitsui#1653
    @requires_toolkit([ToolkitName.qt])
    def test_default_factory(self):
        temp_dir = tempfile.mkdtemp()

        def test_callable():
            return temp_dir

        class Foo(HasStrictTraits):
            dirs = List(Directory(exists=True))
            view = View(
                Item(
                    "dirs",
                    editor=ListEditor(item_factory=test_callable)
                )
            )

        obj = Foo()
        tester = UITester()
        with tester.create_ui(obj) as ui:
            dirs_list_editor = tester.find_by_name(ui, "dirs")
            # should not raise error (see above test_add_item_fails)
            dirs_list_editor._target.add_empty()
            self.assertEqual(len(obj.dirs), 1)

        shutil.rmtree(temp_dir)

    def test_default_factory_with_args(self):
        class Foo(HasStrictTraits):
            bar = Int()
            baz = Str()

        def test_callable(bar, baz=''):
            return Foo(bar=bar, baz=baz)

        class TestFoo(HasStrictTraits):
            foos = List(Foo)
            view = View(
                Item(
                    "foos",
                    editor=ListEditor(
                        item_factory=test_callable,
                        item_factory_args=(7,),
                        item_factory_kwargs={'baz': "python"}
                    )
                )
            )

        obj = TestFoo()
        tester = UITester()
        with tester.create_ui(obj) as ui:
            foos_list_editor = tester.find_by_name(ui, "foos")
            # should not raise error (see above test_add_item_fails)
            foos_list_editor._target.add_empty()
            self.assertEqual(len(obj.foos), 1)
            self.assertEqual(obj.foos[0].bar, 7)
            self.assertEqual(obj.foos[0].baz, "python")


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestNotebookListEditor(unittest.TestCase):

    def test_modify_person_name(self):
        phonebook = Phonebook(
            people=get_people(),
        )
        tester = UITester()
        with tester.create_ui(phonebook, dict(view=notebook_view)) as ui:
            list_ = tester.find_by_name(ui, "people")
            list_.locate(Index(1)).perform(MouseClick())
            name_field = list_.locate(Index(1)).find_by_name("name")
            for _ in range(4):
                name_field.perform(KeyClick("Backspace"))
            name_field.perform(KeySequence("Pete"))

            self.assertEqual(phonebook.people[1].name, "Pete")

    def test_get_person_name(self):
        person1 = Person()
        person2 = Person(name="Mary")
        phonebook = Phonebook(
            people=[person1, person2],
        )
        tester = UITester()
        with tester.create_ui(phonebook, dict(view=notebook_view)) as ui:
            list_ = tester.find_by_name(ui, "people")
            list_.locate(Index(1)).perform(MouseClick())
            name_field = list_.locate(Index(1)).find_by_name("name")
            actual = name_field.inspect(DisplayedText())
            self.assertEqual(actual, "Mary")

    def test_index_out_of_bound(self):
        phonebook = Phonebook(
            people=[],
        )
        tester = UITester()
        with tester.create_ui(phonebook, dict(view=notebook_view)) as ui:
            with self.assertRaises(IndexError):
                tester.find_by_name(ui, "people").\
                    locate(Index(0)).\
                    perform(MouseClick())



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the abstract ViewElement class that all trait view template items
    (i.e., View, Group, Item, Include) derive from.
"""

import re

from traits.api import AbstractViewElement, Bool, HasPrivateTraits, Instance

from .ui_traits import (
    AnObject,
    DockStyle,
    EditorStyle,
    ExportType,
    HelpId,
    Image,
)


label_pat = re.compile(r"^(.*)\[(.*)\](.*)$", re.MULTILINE | re.DOTALL)
label_pat2 = re.compile(r"^(.*){(.*)}(.*)$", re.MULTILINE | re.DOTALL)

# -------------------------------------------------------------------------
#  'ViewElement' class (abstract):
# -------------------------------------------------------------------------


class ViewElement(HasPrivateTraits):
    """ An element of a view.
    """

    def replace_include(self, view_elements):
        """ Searches the current object's **content** attribute for objects that
        have an **id** attribute, and replaces each one with an Include object
        with the same **id** value, and puts the replaced object into the
        specified ViewElements object.

        Parameters
        ----------
        view_elements : ViewElements object
            Object containing Group, Item, and Include objects
        """
        pass  # Normally overridden in a subclass

    def is_includable(self):
        """ Returns whether the object is replacable by an Include object.
        """
        return False  # Normally overridden in a subclass


class DefaultViewElement(ViewElement):
    """ A view element that can be used as a default value for traits whose
        value is a view element.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: The default context object to edit:
    object = AnObject

    #: The default editor style to use:
    style = EditorStyle

    #: The default dock style to use:
    dock = DockStyle

    #: The default notebook tab image to use:
    image = Image

    #: The category of elements dragged out of the view:
    export = ExportType

    #: Should labels be added to items in a group?
    show_labels = Bool(True)


# -------------------------------------------------------------------------
#  Trait definitions:
# -------------------------------------------------------------------------

# The container trait used by ViewSubElements:
Container = Instance(ViewElement, factory=DefaultViewElement)

# -------------------------------------------------------------------------
#  'ViewSubElement' class (abstract):
# -------------------------------------------------------------------------


class ViewSubElement(ViewElement):
    """ Abstract class representing elements that can be contained in a view.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: The object this ViewSubElement is contained in; must be a ViewElement.
    container = Container

    #: External help context identifier:
    help_id = HelpId

    def _split(self, name, value, char, finder, assign, result):
        """ Splits a string at a specified character.
        """
        col = finder(value, char)
        if col < 0:
            return value

        items = (value[:col].strip(), value[col + 1:].strip())
        if items[assign] != "":
            setattr(self, name, items[assign])

        return items[result]

    def _option(self, string, option, name, value):
        """ Sets a object trait if a specified option string is found.
        """
        col = string.find(option)
        if col >= 0:
            string = string[:col] + string[col + len(option):]
            setattr(self, name, value)

        return string

    def _parse_style(self, value):
        """ Parses any of the one-character forms of the **style** trait.
        """
        value = self._option(value, "$", "style", "simple")
        value = self._option(value, "@", "style", "custom")
        value = self._option(value, "*", "style", "text")
        value = self._option(value, "~", "style", "readonly")
        value = self._split("style", value, ";", str.rfind, 1, 0)

        return value

    def _parse_label(self, value):
        """ Parses a '[label]' value from the string definition.
        """
        match = label_pat.match(value)
        if match is not None:
            self._parsed_label()
        else:
            match = label_pat2.match(value)

        empty = False
        if match is not None:
            self.label = match.group(2).strip()
            empty = self.label == ""
            value = match.group(1) + match.group(3)

        return (value, empty)

    def _parsed_label(self):
        """ Handles a label being found in the string definition.
        """
        pass

    def _repr_value(self, value, prefix="", suffix="", ignore=""):
        """ Returns a "pretty print" version of a specified Item trait value.
        """
        if value == ignore:
            return ""

        return "%s%s%s" % (prefix, value, suffix)

    def _repr_options(self, *names):
        """ Returns a 'pretty print' version of a list of traits.
        """
        result = []
        for name in names:
            value = getattr(self, name)
            if value != self.trait(name).default_value_for(self, name):
                result.append((name, repr(value)))

        if len(result) > 0:
            n = max([len(name) for name, value in result])
            return ",\n".join(
                ["%s = %s" % (name.ljust(n), value) for name, value in result]
            )

        return None

    def _indent(self, string, indent="    "):
        """ Indents each line in a specified string by 4 spaces.
        """
        return "\n".join([indent + s for s in string.split("\n")])


# Register ViewElement as implementing AbstractViewElement
# TODO: eventually have ViewElement inherit directly
AbstractViewElement.register(ViewElement)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the editor factory for a drag-and-drop editor. A drag-and-drop
    editor represents its value as a simple image which, depending upon the
    editor style, can be a drag source only, a drop target only, or both a
    drag source and a drop target.
"""

from pyface.ui_traits import Image

from traitsui.editor_factory import EditorFactory


class DNDEditor(EditorFactory):
    """ Editor factory for drag-and-drop editors.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: The image to use for the target:
    image = Image

    #: The image to use when the target is disabled:
    disabled_image = Image


# This alias is deprecated and will be removed in TraitsUI 8.
ToolkitEditorFactory = DNDEditor



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest
from unittest import mock

try:
    from pyface.qt import QtWebkit  # noqa: F401
    NO_WEBKIT_OR_WEBENGINE = False
except ImportError:
    try:
        from pyface.qt import QtWebEngine  # noqa: F401
        NO_WEBKIT_OR_WEBENGINE = False
    except ImportError:
        NO_WEBKIT_OR_WEBENGINE = True
from traits.api import HasTraits, Str
from traitsui.api import HTMLEditor, Item, View
from traitsui.tests._tools import (
    BaseTestMixin,
    is_qt,
    requires_toolkit,
    ToolkitName,
)
from traitsui.testing.api import MouseClick, TargetRegistry, UITester


class HTMLModel(HasTraits):
    """ Dummy class for testing HTMLEditor."""

    content = Str()

    model_base_url = Str()


def get_view(base_url_name):
    return View(
        Item(
            "content",
            editor=HTMLEditor(
                format_text=True,
                base_url_name=base_url_name,
            )
        )
    )


class HTMLContent:
    """ Action to retrieve the HTML content currently displayed.
    Implementation should return a str, whose content conforms to HTML markup.
    """
    pass


def _is_webkit_page(page):
    """ Return true if the given page is a QWebPage from QtWebKit.

    Intended for handling the compatibility between QtWebKit and QtWebEngine.

    Parameters
    ----------
    page : QWebEnginePage or QWebPage
    """
    return hasattr(page, "setLinkDelegationPolicy")


def qt_get_page_html_content(page):
    """ Return the HTML content currently being viewed.

    Parameters
    ----------
    page : QWebEnginePage or QWebPage

    Returns
    -------
    html : str
    """
    qt_allow_view_to_load(page)
    if _is_webkit_page(page):
        return page.mainFrame().toHtml()

    content = []
    page.toHtml(content.append)
    qt_allow_view_to_load(page)
    return "".join(content)


def wait_for_qt_signal(qt_signal, timeout):
    """ Wait for the given Qt signal to fire, or timeout.

    A mock implementation of QSignalSpy.wait, which is one of the missing
    bindings in PySide2, and is not available in Qt4.

    Parameters
    ----------
    qt_signal : signal
        Qt signal to wait for
    timeout : int
        Timeout in milliseconds, to match Qt API.

    Raises
    ------
    RuntimeError
    """
    from pyface.qt import QtCore

    # QEventLoop is used instead of QApplication due to observed
    # hangs with Qt4.
    event_loop = QtCore.QEventLoop()

    def exit(*args, **kwargs):
        event_loop.quit()

    timeout_timer = QtCore.QTimer()
    timeout_timer.setSingleShot(True)
    timeout_timer.setInterval(timeout)
    timeout_timer.timeout.connect(exit)
    qt_signal.connect(exit)

    timeout_timer.start()
    event_loop.exec_()

    qt_signal.disconnect(exit)
    if timeout_timer.isActive():
        timeout_timer.stop()
    else:
        raise RuntimeError("Timeout waiting for signal.")


def qt_allow_view_to_load(loadable, timeout=0.5):
    """ Allow QWebView/QWebPage/QWebEngineView/QWebEnginePage to finish
    loading.

    Out of context, this function does not know if the page has started
    loading. Therefore no timeout error is raised.

    For most testing purposes, this function should be good enough to
    avoid interacting with the Qt web page before it has finished loading, at
    a cost of a slower test.

    Parameters
    ----------
    loadable : QWebView or QWebPage or QWebEngineView or QWebEnginePage
        The view / page to allow loading to finish. Any object with the
        loadFinished signal can be used.
    timeout : float
        Timeout in seconds for each signal being observed.
    """
    timeout_ms = round(timeout * 1000)
    try:
        wait_for_qt_signal(loadable.loadFinished, timeout=timeout_ms)
    except RuntimeError:
        return


def qt_mouse_click_web_view(view, delay):
    """ Perform a mouse click at the center of the web view.

    Note that the page is allowed time to load before and after the mouse
    click.

    Parameters
    ----------
    view : QWebView or QWebEngineView
    """
    from pyface.qt import QtCore
    from pyface.qt.QtTest import QTest

    qt_allow_view_to_load(view)

    if view.focusProxy() is not None:
        # QWebEngineView
        widget = view.focusProxy()
    else:
        # QWebView
        widget = view

    try:
        QTest.mouseClick(
            widget,
            QtCore.Qt.LeftButton,
            QtCore.Qt.NoModifier,
            delay=delay,
        )
    finally:
        qt_allow_view_to_load(view)


def qt_target_registry():
    """ Return an instance of TargetRegistry for testing Qt + HTMLEditor

    Returns
    -------
    target_registry : TargetRegistry
    """
    from traitsui.qt4.html_editor import SimpleEditor

    registry = TargetRegistry()
    registry.register_interaction(
        target_class=SimpleEditor,
        interaction_class=MouseClick,
        handler=lambda wrapper, _: qt_mouse_click_web_view(
            wrapper._target.control, wrapper.delay
        )
    )
    registry.register_interaction(
        target_class=SimpleEditor,
        interaction_class=HTMLContent,
        handler=lambda wrapper, _: (
            qt_get_page_html_content(wrapper._target.control.page())
        )
    )
    return registry


def get_custom_ui_tester():
    """ Return an instance of UITester that contains extended testing
    functionality for HTMLEditor. These implementations are used by
    TraitsUI only, are more ad hoc than they would have been if they were made
    public.
    """
    if is_qt():
        return UITester(registries=[qt_target_registry()])
    return UITester()


# Run this against wx as well once enthought/traitsui#752 is fixed.
@requires_toolkit([ToolkitName.qt])
@unittest.skipIf(
    NO_WEBKIT_OR_WEBENGINE, "Tests require either QtWebKit or QtWebEngine"
)
class TestHTMLEditor(BaseTestMixin, unittest.TestCase):
    """ Test HTMLEditor """

    def setUp(self):
        BaseTestMixin.setUp(self)
        self.tester = get_custom_ui_tester()

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def test_init_and_dispose(self):
        # Smoke test to check init and dispose do not fail.
        model = HTMLModel()
        view = get_view(base_url_name="")
        with self.tester.create_ui(model, dict(view=view)):
            pass

    def test_base_url_changed(self):
        # Test if the base_url is changed after the UI closes, nothing
        # fails because sync_value is unhooked in the base class.
        model = HTMLModel()
        view = get_view(base_url_name="model_base_url")
        with self.tester.create_ui(model, dict(view=view)):
            pass
        # It is okay to modify base_url after the UI is closed
        model.model_base_url = "/new_dir"

    @requires_toolkit([ToolkitName.qt])
    def test_open_internal_link(self):
        # this test requires Qt because it relies on the link filling up
        # the entire page through the use of CSS, which isn't supported
        # by Wx.
        model = HTMLModel(content="""
        <html>
            <a
              href='/#'
              target='_self'
              style='display:block; width: 100%; height: 100%'>
                Internal Link
            </a>
        </html>
        """)
        view = View(
            Item("content", editor=HTMLEditor())
        )

        with self.tester.create_ui(model, dict(view=view)) as ui:
            html_view = self.tester.find_by_name(ui, "content")

            # when
            with mock.patch("webbrowser.open_new") as mocked_browser:
                html_view.perform(MouseClick())

        mocked_browser.assert_not_called()

    @requires_toolkit([ToolkitName.qt])
    def test_open_external_link(self):
        # this test requires Qt because it relies on the link filling up
        # the entire page through the use of CSS, which isn't supported
        # by Wx.
        model = HTMLModel(content="""
        <html>
            <a
              href='test://testing'
              target='_blank'
              style='display:block; width: 100%; height: 100%'>
                External Link
            </a>
        </html>
        """)
        view = View(
            Item("content", editor=HTMLEditor())
        )

        with self.tester.create_ui(model, dict(view=view)) as ui:
            html_view = self.tester.find_by_name(ui, "content")
            with mock.patch("webbrowser.open_new") as mocked_browser:
                html_view.perform(MouseClick())
            self.assertIn(
                "External Link",
                html_view.inspect(HTMLContent()),
            )

        # See enthought/traitsui#1464
        # This is the expected behaviour:
        # mocked_browser.assert_called_once_with("test://testing")
        # However, this is the current unexpected behaviour
        mocked_browser.assert_not_called()

    @requires_toolkit([ToolkitName.qt])
    def test_open_internal_link_externally(self):
        # this test requires Qt because it relies on the link filling up
        # the entire page through the use of CSS, which isn't supported
        # by Wx.
        model = HTMLModel(content="""
        <html>
            <a
              href='test://testing'
              target='_self'
              style='display:block; width: 100%; height: 100%'>
                Internal Link
            </a>
        </html>
        """)
        view = View(
            Item("content", editor=HTMLEditor(open_externally=True))
        )

        with self.tester.create_ui(model, dict(view=view)) as ui:
            html_view = self.tester.find_by_name(ui, "content")
            with mock.patch("webbrowser.open_new") as mocked_browser:
                html_view.perform(MouseClick())
            self.assertIn(
                "Internal Link",
                html_view.inspect(HTMLContent()),
            )

        mocked_browser.assert_called_once_with("test://testing")

    @requires_toolkit([ToolkitName.qt])
    def test_open_external_link_externally(self):
        model = HTMLModel(content="""
        <html>
            <a
              href='test://testing'
              target='_blank'
              style='display:block; width: 100%; height: 100%'>
                External Link
            </a>
        </html>
        """)
        view = View(
            Item("content", editor=HTMLEditor(open_externally=True))
        )

        with self.tester.create_ui(model, dict(view=view)) as ui:
            html_view = self.tester.find_by_name(ui, "content")
            with mock.patch("webbrowser.open_new") as mocked_browser:
                html_view.perform(MouseClick())
            self.assertIn(
                "External Link",
                html_view.inspect(HTMLContent()),
            )

            is_webkit = _is_webkit_page(html_view._target.control.page())

        if is_webkit:
            # This is the expected behavior.
            mocked_browser.assert_called_once_with("test://testing")
        else:
            # Expected failure:
            # See enthought/traitsui#1464
            # This is the current unexpected behavior if QtWebEngine is used.
            mocked_browser.assert_not_called()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the various text editors for the Qt user interface toolkit.
    The module is mainly a place-folder for TextEditor factories that have
    been augmented to also listen to changes in the items of the list object.
"""


from .text_editor import SimpleEditor as QtSimpleEditor
from .text_editor import CustomEditor as QtCustomEditor
from .text_editor import ReadonlyEditor as QtReadonlyEditor
from ..editors.csv_list_editor import _prepare_method, _dispose_method


class SimpleEditor(QtSimpleEditor):
    """ Simple Editor style for CSVListEditor. """

    prepare = _prepare_method
    dispose = _dispose_method


class CustomEditor(QtCustomEditor):
    """ Custom Editor style for CSVListEditor. """

    prepare = _prepare_method
    dispose = _dispose_method


class ReadonlyEditor(QtReadonlyEditor):
    """ Readonly Editor style for CSVListEditor. """

    prepare = _prepare_method
    dispose = _dispose_method


TextEditor = SimpleEditor



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" This module contains helper functions for working with layouts inside an
editor. For example, converting indices so they count through the layout
appropriately.
"""


def column_major_to_row_major(index, n, num_rows, num_cols):
    """ Helper function to convert an index of a grid layout so that the
    index counts over the grid in the correct direction.
    In TraitsUI, grids are typically populated in row major order, however,
    the elements can be assigned to each entry in the grid so that when
    displayed they appear in column major order. To access the correct element
    we may need to convert a column-major based index into a row-major one.

    Parameters
    ----------
    index : int
        the index of interest
    n : int
        The total number of elements in the layout
    num_rows : int
        the number of rows in the layout
    num_cols : int
        the number of columns in the layout

    Returns
    -------
    int
        The converted index (now corresponding to row_major order)

    Notes
    -----
    Since elements are populated in row major order, the resulting grid ends
    up having at most its last row as incomplete. The general approach for the
    algorithm is to find the coordinates as if we had counted through the
    grid in column major order, and then convert that back to a row major
    index. The complications come from the fact that the last row may be
    missing entries.
    Consider the example (n=17, num_row=4, num_cols=5)
    0  4  8  11  14
    1  5  9  12  15
    2  6  10 13  16
    3  7  /  /  /
    Row major order is 0, 4, 11, 14, 1, 5, ...

    If the given index is 7 then we are in a column of the matrix where
    all rows are full.  This corresponds to the else branch below. From here,
    we simply find the (i,j) coordinates of the entry 7 above, with the upper
    left corner representing (0,0).  Thus, (i,j) = (3,1).  Now, to convert
    this to a row major index, we can simply take i * num_cols + j.

    If the given index is 15, then we are in a column where the last row is
    missing an entry.  See the if branch below.  In the case the logic used
    before won't work, because it would expect the grid to be filled. (i.e.
    if one tried to do this in the same way you would get (i,j) = (3,4)).
    To adderess this we break up the grid into two grids:
    0  4     and    8  11  14
    1  5            9  12  15
    2  6            10 13  16
    3  7
    we find the (i2,j2) coordinates of the entry 15 above in grid2, with the
    upper left corner of grid2 representing (0,0). Hence, (i2,j2) = (1,2).
    We then find that index if we had counted in row major order for grid2,
    which would be i2 * num_empty_entries_last_row + j2 = 5, and add that to
    however many elements would need to be counted over from grid one to reach
    element 15 if we had been counting in row major order. Which is
    (num_cols - num_empty_entries_last_row)*(i2+1).
    """
    if index > n:
        raise ValueError("Index is higher number of elements in layout")
    num_empty_entries_last_row = num_cols * num_rows - n
    if num_empty_entries_last_row < 0:
        raise ValueError("n can not be greater than num_cols * num_rows")

    if index > num_rows * (num_cols - num_empty_entries_last_row):
        num_entries_grid1 = num_rows * (num_cols - num_empty_entries_last_row)
        new_index = index - num_entries_grid1
        i2 = new_index % (num_rows - 1)
        j2 = new_index // (num_rows - 1)
        return (num_cols - num_empty_entries_last_row)*(i2+1) + \
            (i2 * num_empty_entries_last_row + j2)
    else:
        i = index % num_rows
        j = index // num_rows
    return i * num_cols + j



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the concrete implementations of the traits Toolkit interface for
    the wxPython user interface toolkit.
"""

# -------------------------------------------------------------------------
#  Define the reference to the exported GUIToolkit object:
# -------------------------------------------------------------------------

from . import toolkit

# Reference to the GUIToolkit object for wxPython
toolkit = toolkit.GUIToolkit("traitsui", "wx", "traitsui.wx")



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines a completely empty editor, intended to be used as a spacer.
"""


from pyface.qt import QtGui

from .editor import Editor


class NullEditor(Editor):
    """ A completely empty editor.
    """

    def init(self, parent):
        """ Finishes initializing the editor by creating the underlying toolkit
            widget.
        """
        self.control = QtGui.QWidget()

    def update_editor(self):
        """ Updates the editor when the object trait changes externally to the
            editor.
        """
        pass



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines constants used by the wxPython implementation of the various text
    editors and text editor factories.
"""

import sys

import wx

from pyface.api import SystemMetrics

#: Define platform and wx version constants:
is_mac = sys.platform == "darwin"

#: Default dialog title
DefaultTitle = "Edit properties"

#: Color of valid input
OKColor = wx.WHITE

#: Color to highlight input errors
ErrorColor = wx.Colour(255, 192, 192)

#: Color for background of read-only fields
ReadonlyColor = wx.Colour(244, 243, 238)

#: Color for background of fields where objects can be dropped
DropColor = wx.Colour(215, 242, 255)

#: Color for an editable field
EditableColor = wx.WHITE

#: Color for background of windows (like dialog background color)
if is_mac:
    WindowColor = wx.Colour(232, 232, 232)
    BorderedGroupColor = wx.Colour(224, 224, 224)
else:
    WindowColor = wx.SystemSettings.GetColour(wx.SYS_COLOUR_MENUBAR)

#: Default colour for table foreground
TableCellColor = wx.BLACK

#: Default colour for table background
TableCellBackgroundColor = wx.WHITE

#: Default colour for table background
TableReadOnlyBackgroundColor = ReadonlyColor

#: Default colour for table background
TableLabelColor = TableCellColor

#: Default colour for table background
TableLabelBackgroundColor = WindowColor

#: Default foreground colour for table selection
TableSelectionTextColor = TableCellColor

#: Default background colour for table selection (light blue)
TableSelectionBackgroundColor = wx.Colour(173, 216, 230)

#: Standard width of an image bitmap
standard_bitmap_width = 120

#: Width of a scrollbar
scrollbar_dx = wx.SystemSettings.GetMetric(wx.SYS_VSCROLL_X)

#: Screen width
screen_dx = SystemMetrics().screen_width

#: Screen height
screen_dy = SystemMetrics().screen_height



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" This file provides tests for the demo of the same name located in the
directory one level up.  
"""
import os
import runpy
import unittest

#: Filename of the demo script
FILENAME = "Image_editor_demo.py"

#: Path of the demo script
DEMO_PATH = os.path.join(os.path.dirname(__file__), "..", FILENAME)


class TestImageEditorDemo(unittest.TestCase):

    def test_image_path_exists(self):
        search_path, = runpy.run_path(DEMO_PATH)["search_path"]
        self.assertTrue(os.path.exists(search_path))


# Run the test(s)
unittest.TextTestRunner().run(
    unittest.TestLoader().loadTestsFromTestCase(TestImageEditorDemo)
)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
**WARNING**

  This demo might not work as expected and some documented features might be
  missing.

-------------------------------------------------------------------------------

Implementation of an ImageEnumEditor demo plugin for the Traits UI demo
program.

This demo shows each of the four styles of the ImageEnumEditor.

Please refer to the `ImageEnumEditor API docs`_ for further information.

.. _ImageEnumEditor API docs: https://docs.enthought.com/traitsui/api/traitsui.editors.image_enum_editor.html#traitsui.editors.image_enum_editor.ImageEnumEditor
"""
# Issues related to the demo warning:
# enthought/traitsui#947


from traits.api import Enum, HasTraits, Str

from traitsui.api import Item, Group, View, ImageEnumEditor

# This list of image names (with the standard suffix "_origin") is used to
# construct an image enumeration trait to demonstrate the ImageEnumEditor:
image_list = ['top left', 'top right', 'bottom left', 'bottom right']


class Dummy(HasTraits):
    """ Dummy class for ImageEnumEditor
    """
    x = Str()

    traits_view = View()


class ImageEnumEditorDemo(HasTraits):
    """ Defines the ImageEnumEditor demo class.
    """

    # Define a trait to view:
    image_from_list = Enum(
        *image_list,
        editor=ImageEnumEditor(
            values=image_list,
            prefix='@icons:',
            suffix='_origin',
            cols=4,
            klass=Dummy
        )
    )

    # Items are used to define the demo display, one Item per editor style:
    img_group = Group(
        Item('image_from_list', style='simple', label='Simple'),
        Item('_'),
        Item('image_from_list', style='text', label='Text'),
        Item('_'),
        Item('image_from_list', style='readonly', label='ReadOnly'),
        Item('_'),
        Item('image_from_list', style='custom', label='Custom')
    )

    # Demo view:
    traits_view = View(
        img_group,
        title='ImageEnumEditor',
        buttons=['OK'],
        resizable=True
    )


# Create the demo:
demo = ImageEnumEditorDemo()

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Implement the interface of AbstractTargetRegistry to support
dynamic checking on the target.
"""
import inspect

from traitsui.testing.tester._abstract_target_registry import (
    AbstractTargetRegistry,
)
from traitsui.testing.tester.exceptions import (
    InteractionNotSupported,
    LocationNotSupported,
)


class DynamicTargetRegistry(AbstractTargetRegistry):
    """ Registry to support testing targets that satisfy a given criterion.

    An instance of this registry can be used with ``UITester`` and
    ``UIWrapper`` such that all given interactions and handlers will be
    applicable to any target that deemed to be supported by ``can_support``.
    The general priority rule applies. See :ref:`testing-how-extension-works`
    in the User Manual for further details.

    For stricter checks on the target type, use
    :class:`~traitsui.testing.tester.target_registry.TargetRegistry`.

    As an example, this registry::

        registry = DynamicTargetRegistry(
            can_support=lambda target: target.__class__ is MyEditorClass,
            interaction_to_handler={MyAction: handler},
        )

    has equivalent behaviors compared to this registry::

        registry = TargetRegistry()
        registry.register_interaction(MyEditorClass, MyAction, handler)

    Parameters
    ----------
    can_support : callable(target) -> bool
        A callable that return true if the given target is supported
        by the registry.
    interaction_to_handler : dict(type, callable)
        A dictionary mapping from interaction type to handler callables.
        Each handler callable in the values should have the signature
        ``callable(UIWrapper, interaction) -> any``, where ``instance`` should
        have the type defined in the associated key.
    """

    def __init__(self, *, can_support, interaction_to_handler):
        self.can_support = can_support
        self.interaction_to_handler = interaction_to_handler

    def _get_handler(self, target, interaction):
        """ Return a callable for handling an interaction for a given target.

        This is an implementation for an abstract method.

        Parameters
        ----------
        target : any
            The UI target being operated on.
        interaction : any
            Any interaction object.

        Returns
        -------
        handler : callable(UIWrapper, interaction) -> any
            The function to handle the given interaction on a target.

        Raises
        ------
        InteractionNotSupported
            If the given target and interaction types are not supported
            by this registry.
        """
        if interaction.__class__ not in self._get_interactions(target):
            raise InteractionNotSupported(
                target_class=target.__class__,
                interaction_class=interaction.__class__,
                supported=list(self._get_interactions(target)),
            )
        return self.interaction_to_handler[interaction.__class__]

    def _get_interactions(self, target):
        """ Returns all the interactions supported for the given target.

        This is an implementation for an abstract method.

        Parameters
        ----------
        target : any
            The UI target for which supported interactions are queried.

        Returns
        -------
        interaction_classes : set
            Supported interaction types for the given target type.
        """
        if self.can_support(target):
            return set(self.interaction_to_handler)
        return set()

    def _get_interaction_doc(self, target, interaction_class):
        """ Return the documentation for the given target and interaction type.

        This is an implementation for an abstract method.

        Parameters
        ----------
        target : any
            The UI target for which the interaction will be applied.
        interaction_class : subclass of type
            Any class.

        Returns
        -------
        doc : str

        Raises
        ------
        InteractionNotSupported
            If the given target and interaction types are not supported
            by this registry.
        """
        if interaction_class not in self._get_interactions(target):
            raise InteractionNotSupported(
                target_class=target.__class__,
                interaction_class=interaction_class,
                supported=list(self._get_interactions(target)),
            )
        return inspect.getdoc(interaction_class)

    def _get_solver(self, target, location):
        """ Return a callable registered for resolving a location for the
        given target and location.

        This is an implementation for an abstract method.

        Parameters
        ----------
        target : any
            The UI target being operated on.
        location : subclass of type
            The location to be resolved on the target.

        Raises
        ------
        LocationNotSupported
            If the given locator and target types are not supported.
        """
        raise LocationNotSupported(
            target_class=target.__class__,
            locator_class=location.__class__,
            supported=[],
        )

    def _get_locations(self, target):
        """ Returns all the location types supported for the given target.

        This is an implementation for an abstract method.

        Parameters
        ----------
        target : any
            The UI target for which supported location types are queried.

        Returns
        -------
        locators_classes : set
            Supported locator types for the given target type.
        """
        return set()

    def _get_location_doc(self, target, locator_class):
        """ Return the documentation for the given target and locator type.

        This is an implementation for an abstract method.

        Parameters
        ----------
        target : any
            The UI target being operated on.
        locator_class : subclass of type
            Any class.

        Returns
        -------
        doc : str

        Raises
        ------
        LocationNotSupported
            If the given locator and target types are not supported.
        """
        raise LocationNotSupported(
            target_class=target.__class__,
            locator_class=locator_class,
            supported=[],
        )



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Trait definition for a null-based (i.e., no UI) font.
"""



from traits.api import Trait, TraitHandler, TraitError

# ------------------------------------------------------------------------------
#  Convert a string into a valid 'wxFont' object (if possible):
# ------------------------------------------------------------------------------

# Mapping of strings to valid wxFont families
font_families = ["default", "decorative", "roman", "script", "swiss", "modern"]

# Mapping of strings to wxFont styles
font_styles = ["slant", "italic"]

# Mapping of strings wxFont weights
font_weights = ["light", "bold"]

# Strings to ignore in text representations of fonts
font_noise = ["pt", "point", "family"]


# ------------------------------------------------------------------------------
#  'TraitFont' class'
# ------------------------------------------------------------------------------


class TraitFont(TraitHandler):
    """ Ensures that values assigned to a trait attribute are valid font
    descriptor strings; the value actually assigned is the corresponding
    canonical font descriptor string.
    """

    def validate(self, object, name, value):
        """ Validates that the value is a valid font descriptor string.
        """
        try:
            point_size = family = style = weight = underline = ""
            facename = [""]
            for word in value.split():
                lword = word.lower()
                if lword in font_families:
                    family = " " + lword
                elif lword in font_styles:
                    style = " " + lword
                elif lword in font_weights:
                    weight = " " + lword
                elif lword == "underline":
                    underline = " " + lword
                elif lword not in font_noise:
                    try:
                        int(lword)
                        point_size = lword + " pt"
                    except:
                        facename.append(word)
            fontstr = (
                "%s%s%s%s%s%s"
                % (
                    point_size,
                    family,
                    style,
                    weight,
                    underline,
                    " ".join(facename),
                )
            ).strip()
            return fontstr
        except Exception:
            pass
        raise TraitError(object, name, "a font descriptor string", repr(value))

    def info(self):
        return (
            "a string describing a font (e.g. '12 pt bold italic "
            "swiss family Arial' or 'default 12')"
        )


# ------------------------------------------------------------------------------
#  Define a 'null' specific font trait:
# ------------------------------------------------------------------------------

### Note: Declare the editor to be a function which returns the FontEditor
# class from traits ui to avoid circular import issues. For backwards
# compatibility with previous Traits versions, the 'editors' folder in Traits
# project declares 'from api import *' in its __init__.py. The 'api' in turn
# can contain classes that have a Font trait which lead to this file getting
# imported. This leads to a circular import when declaring a Font trait.
def get_font_editor(*args, **traits):
    from ..api import FontEditor

    return FontEditor(*args, **traits)


fh = TraitFont()
NullFont = Trait(
    fh.validate(None, None, "Arial 10"), fh, editor=get_font_editor
)



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# multi_object_view.py -- Sample code to show multi-object view with context

from traits.api import Bool, HasTraits, Int, Str
from traitsui.api import Group, Item, View


class House(HasTraits):
    address = Str()
    bedrooms = Int()
    pool = Bool()
    price = Int()


# View object designed to display two objects of class 'House'
comp_view = View(
    Group(
        Group(
            Item('h1.address', resizable=True),
            Item('h1.bedrooms'),
            Item('h1.pool'),
            Item('h1.price'),
            show_border=True,
        ),
        Group(
            Item('h2.address', resizable=True),
            Item('h2.bedrooms'),
            Item('h2.pool'),
            Item('h2.price'),
            show_border=True,
        ),
        orientation='horizontal',
    ),
    title='House Comparison',
)

# A pair of houses to demonstrate the View
house1 = House(
    address='4743 Dudley Lane',
    bedrooms=3,
    pool=False,
    price=150000,
)
house2 = House(
    address='11604 Autumn Ridge',
    bedrooms=3,
    pool=True,
    price=200000,
)

# ...And the actual display command
house1.configure_traits(view=comp_view, context={'h1': house1, 'h2': house2})



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
This demo shows you how to use animated GIF files in a traits user interface.
Note that this demo only works with wx backend.
"""

from os.path import join, dirname, abspath

from traits.api import HasTraits, File, Bool

from traitsui.api import View, VGroup, HGroup, Item, EnumEditor

from traitsui.wx.animated_gif_editor import AnimatedGIFEditor


# Some sample animated GIF files:
base_path = join(dirname(__file__), 'images')

files = [
    abspath(join(base_path, 'logo_64x64.gif')),
    abspath(join(base_path, 'logo_48x48.gif')),
    abspath(join(base_path, 'logo_32x32.gif'))
]


class AnimatedGIFDemo(HasTraits):

    # The animated GIF file to display:
    gif_file = File(files[0])

    # Is the animation playing or not?
    playing = Bool(True)

    # The traits view:
    view = View(
        VGroup(
            HGroup(
                Item('gif_file',
                     editor=AnimatedGIFEditor(playing='playing'),
                     show_label=False),
                Item('playing'),
            ),
            '_',
            Item('gif_file',
                 label='GIF File',
                 editor=EnumEditor(values=files)
                 )
        ),
        title='Animated GIF Demo',
        buttons=['OK']
    )

# Create the demo:
demo = AnimatedGIFDemo()

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2008-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# ------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms
# described in the PyQt GPL exception also apply

#
# Author: Riverbank Computing Limited
# ------------------------------------------------------------------------------

""" Defines the tuple editor for the PyQt user interface toolkit.
"""


from traitsui.editors.tuple_editor import SimpleEditor as BaseSimpleEditor

from .editor import Editor

# -------------------------------------------------------------------------
#  'SimpleEditor' class:
# -------------------------------------------------------------------------


class SimpleEditor(BaseSimpleEditor, Editor):
    """ Simple style of editor for tuples.

    The editor displays an editor for each of the fields in the tuple, based on
    the type of each field.
    """

    pass



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Define top-level object(s) to support testing TraitsUI applications.
"""

import contextlib

from traitsui.testing._gui import process_cascade_events
from traitsui.testing._exception_handling import reraise_exceptions
from traitsui.testing.tester._ui_tester_registry.default_registry import (
    get_default_registries
)
from traitsui.testing.tester.ui_wrapper import UIWrapper


class UITester:
    """ UITester assists testing of GUI applications developed using TraitsUI.

    See :ref:`testing-traitsui-applications` Section in the User Manual for
    further details.

    Parameters
    ----------
    registries : list of TargetRegistry, optional
        Registries of interaction for different targets, in the order
        of decreasing priority. If provided, a shallow copy will be made.
        Additional registries are appended to the list to provide
        builtin support for TraitsUI UI and editors.
    delay : int, optional
        Time delay (in ms) in which actions by the tester are performed. Note
        it is propagated through to created child wrappers. The delay allows
        visual confirmation test code is working as desired. Defaults to 0.
    auto_process_events : bool, optional
        Whether to process (cascade) GUI events automatically. Default is True.
        For tests that launch a modal dialog and rely on a recurring timer to
        poll if the dialog is closed, it may be necessary to set this flag to
        false in order to avoid deadlocks. Note that this is propagated
        through to created child wrappers.

    Attributes
    ----------
    delay : int
        Time delay (in ms) in which actions by the tester are performed. Note
        it is propagated through to created child wrappers. The delay allows
        visual confirmation test code is working as desired.
    """

    def __init__(self, *, registries=None, delay=0, auto_process_events=True):
        if registries is None:
            self._registries = []
        else:
            self._registries = registries.copy()

        # These registries contribute the support for TraitsUI UI and editors.
        # The find_by_name/find_by_id methods in this class also depend on
        # one of these registries.
        self._registries.extend(get_default_registries())
        self.delay = delay
        self._auto_process_events = auto_process_events

    @property
    def auto_process_events(self):
        """ Flag to indicate whether to process (cascade) GUI events
        automatically. This is propagated through to
        :class:`~traitsui.testing.tester.ui_wrapper.UIWrapper` instances
        created from this tester.

        This value can only be set at instantiation.
        """
        # No mutations post-init are allowed because UIWrapper created prior to
        # the mutation will not respect the new value, and so it would be
        # confusing if this attribute was allowed to be mutated.
        return self._auto_process_events

    @contextlib.contextmanager
    def create_ui(self, object, ui_kwargs=None):
        """ Context manager to create a UI and dispose it upon exit.

        This method does not modify the states of the :class:`UITester` and is
        not a requirement for using other methods on the tester.

        Parameters
        ----------
        object : HasTraits
            An instance of HasTraits for which a GUI will be created.
        ui_kwargs : dict or None, optional
            Keyword arguments to be provided to ``HasTraits.edit_traits``.
            Default is to call ``edit_traits`` with no additional keyword
            arguments.

        Yields
        ------
        ui : traitsui.ui.UI
        """

        ui_kwargs = {} if ui_kwargs is None else ui_kwargs
        ui = object.edit_traits(**ui_kwargs)
        try:
            yield ui
        finally:
            with reraise_exceptions():
                if self._auto_process_events:
                    # At the end of a test, there may be events to be
                    # processed. If dispose happens first, those events will be
                    # processed after various editor states are removed,
                    # causing errors. But if we are asked not to process
                    # events, then don't.
                    process_cascade_events()
                try:
                    ui.dispose()
                finally:
                    # dispose is not atomic and may push more events to the
                    # event queue. Flush those too unless we are asked not to.
                    if self._auto_process_events:
                        process_cascade_events()

    def find_by_name(self, ui, name):
        """ Find the TraitsUI editor with the given name and return a new
        ``UIWrapper`` object for further interactions with the editor.

        ``name`` is typically a value defined on
        :attr:`traitsui.item.Item.name`. This name is passed onto
        :func:`traitsui.ui.UI.get_editors`.

        Parameters
        ----------
        ui : traitsui.ui.UI
            The UI created, e.g. by ``create_ui``.
        name : str
            A single name for retrieving a target on a UI.

        Returns
        -------
        wrapper : UIWrapper
        """
        return self._get_wrapper(ui).find_by_name(name=name)

    def find_by_id(self, ui, id):
        """ Find the TraitsUI editor with the given identifier and return a new
        ``UIWrapper`` object for further interactions with the editor.

        ``id`` is typically a value defined on :attr:`traitsui.item.Item.id`
        or :attr:`traitsui.group.Group.id`. This provides a shortcut to locate
        a very nested editor in a view.

        Parameters
        ----------
        ui : traitsui.ui.UI
            The UI created, e.g. by ``create_ui``.
        id : str
            Id for finding an item in the UI.

        Returns
        -------
        wrapper : UIWrapper
        """
        return self._get_wrapper(ui).find_by_id(id=id)

    def _get_wrapper(self, ui):
        """ Return a new UIWrapper wrapping the given traitsui.ui.UI.

        Parameters
        ----------
        ui : traitsui.ui.UI
            The UI created, e.g. by ``create_ui``.

        Returns
        -------
        wrapper : UIWrapper
        """
        return UIWrapper(
            target=ui,
            registries=self._registries,
            delay=self.delay,
            auto_process_events=self._auto_process_events,
        )



# (C) Copyright 2004-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the progress editor factory for all traits toolkit backends,
"""

from traits.api import Int, Bool, Str

from traitsui.editor_factory import EditorFactory


class ProgressEditor(EditorFactory):
    """ Editor factory for code editors.
    """

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: The title
    title = Str()

    #: The message to be displayed along side the progress guage
    message = Str()

    #: The name of an [object.]trait that defines the message string
    message_name = Str()

    #: The starting value
    min = Int()

    #: The name of an [object.]trait that defines the starting value
    min_name = Str()

    #: The ending value
    max = Int()

    #: The name of an [object.]trait that defines the ending value
    max_name = Str()

    #: If the cancel button should be shown (not very sensible as an editor)
    can_cancel = Bool(False)

    #: If the estimated time should be shown (not very sensible as an editor)
    show_time = Bool(False)

    #: if the percent complete should be shown
    show_percent = Bool(False)


# This alias is deprecated and will be removed in TraitsUI 8.
ToolkitEditorFactory = ProgressEditor
